# Eval.pm
# implements 'eval', 'reeval' and 'archive' commands
#
# v1.19, 2004-01-15
# Copyright (c) Michel Messerschmidt 2002-2004
#
package Avtc::Eval;
use strict;
use Exporter;
use File::Copy;
use File::Basename;
use Cwd;
use IO::File;
use Avtc::Evalcore;   # contains core functions for most evaluation steps
use Avtc::Eval_chk;   # contains evaluation consistency check functions

@Avtc::Eval::ISA = qw(Exporter);
@Avtc::Eval::EXPORT = qw(archive evaluate reeval);


sub archive {
    $::archiveOnly = 1;
    my $ret = doEval();
    if ($ret) {
        ::log("ERROR evaluating logfiles\n", 0);
    }
    ::log("Executing command: status (". $::cmds{"status"}->{"desc"} .")\n\n", 0);
    $ret += ::reportStatus();
    if ($ret) {
        ::log("ERROR creating status report\n", 0);
    }
    return $ret;
}

# the name 'eval' is not a good choice here :-)
sub evaluate {
    $::reEval = 0;
    my $ret = doEval();
    if ($ret) {
        ::log("ERROR evaluating logfiles\n", 0);
    }
    return $ret;
}

sub reeval {
    $::reEval = 1;
    my $ret = doEval();
    if ($ret) {
        ::log("ERROR evaluating logfiles\n", 0);
    }
    return $ret;
}


sub doEval {
    # '0' indicates success, everything else an error
    my $ret = 0;
    # exitstatus from system calls is 0, when call was successful
    my $error = 0;

    # archiveOnly implies skipReload = true and reEval = false
    # this is useful to avoid loading big files for nothing
    # (or many conditions in SWITCH statements)
    if ($::archiveOnly) {
        $::skipReload = 1;
        $::reEval = 0;
    }

    # now we have all needed information to get the data from the scanner logs
    #begin the big loop over all OSs, TBs and Products
    foreach my $os (sort keys %::osList) {

    LOOP_TB: foreach my $tb (sort keys %::lists) {
        # now we can define the complete paths to operate on
        my $EvalDir = $::EvalBase . $::sep . $os . $::sep . $tb;
        my $LogDir = $::LogBase . $::sep . $os . $::sep . $tb;
        my $NewLogDir = $::NewLogBase . $::sep . $os . $::sep .
                         $::lists{$tb}->{"id"};
        my $TempDir = $::TempBase . $::sep . $os . $::sep . $tb;
        my $excludeRef = \@{$::lists{$tb}->{"excluded"}};
        # This MUST be created in two steps to avoid creation of the
        # sub-array "packed" on unpacked testbeds
        my @tbPackList;
        if (exists $::lists{$tb}->{"packed"}) {
            @tbPackList = keys %{$::lists{$tb}->{"packed"}->{"prg"}};
        }
        if ($::debug > 1) {
            ::log("EvalDir = $EvalDir\n", 2);
            ::log("LogDir = $LogDir\n", 2);
            ::log("NewLogDir = $NewLogDir\n", 2);
            ::log("TempDir = $TempDir\n", 2);
            ::log("excList for $tb:\n", 2);
            foreach (@{$excludeRef}) {
                ::log($_."\n", 4);
            }
        }

        # check existance of TempDir and create it if nessecary
        ::makeDir($TempDir) or do {
            ::log("ERROR: creating directory $TempDir\n", 1);
        };
        # change to TempDir, so all following IO can operate on current dir
        chdir($TempDir) or do {
            ::log("ERROR: chdir $TempDir failed:  $!\n", 1);
            ::log("!!! Evaluation of $tb stopped !!!\n\n", 2);
            $ret = 1;
            next LOOP_TB;
        };
        if ($::debug) {
            ::log("Current dir set to ". cwd() ."\n", 1);
        }

        # delete all old files from tempdir
        ::deleteTempFiles($TempDir, "files", 0);

        # test if all evaluation lists exist to avoid false results
        my $file;
        foreach $file (keys %{$::{$tb}->{"evals"}}) {
            if (not -e $::ListDir.$::sep.$file) {
                ::log("ERROR: file not found: ".$::ListDir.$::sep.$file.
                      "\n", 1);
                ::log("!!! Evaluation of $tb stopped !!!\n\n", 2);
                $ret = 1;
                next LOOP_TB;
            }
        }

        my $prod;
        LOOP_PROD: foreach $prod (sort keys %::prodList) {
            # check for not tested combinations
            foreach (@::noTestListAll) {
                if ($_ eq "$os $tb $prod") {
                    if ($::debug) {
                        ::log("not tested: $os - $tb - $prod\n", 1);
                    }
                    next LOOP_PROD;
                }
            }
            ::log("processing $prod - $tb - $os\n", 1);
            # define name of product related files
            my $statFile = $EvalDir . $::sep . $prod . $::FileExtStatus;
            my $checkFile = $prod . "." . $::FileExtControl;
            my $ful = $prod . "." . $::FileExtReport;

            # get status if current product
            # status is one of (see status.txt for more details):
            #   NO ...  = no evaluation for some reason
            #             (eg. no report created or not tested under this OS)
            #   OK ...  = test finished for this prod
            #   -S1     = no scan (no logfile found), no eval / QA
            #   S1      = scanned (ful moved from NewLogDir to LogDir),
            #             but no eval, no QA yet
            #   S1E     = scanned, evaluated, no QA ( by hand)
            #   -S2     = S1E + no 2.scan, no 2.eval, no 2.QA
            #   S2      = S1E + 2.scan, no 2.eval, no 2.QA
            #   S2E     = S1E + 2.scan, 2.eval, no 2.QA
            #   -S3     = S2E + no 3.scan, no 3.eval, no 3.QA
            #   S3      = S2E + 3.scan, no 3.eval, no 3.QA
            #   S3E     = S2E + 3.scan, 3.eval, no 3.QA
            my $status = "-S1";
            if (not -e $statFile) {
                if ($::debug) {
                    ::log("no status file found for $prod\n", 2);
                }
                ::updateStatus($statFile, $status) and do {
                    ::log("ERROR writing status file: $!\n", 2);
                    ::log("!!! Evaluation of $prod stopped !!!\n\n", 2);
                    $ret = 1;
                    next LOOP_PROD;
                };
            } elsif (-z $statFile) {
                ::log("ERROR: empty statusfile $statFile - skipping ".$prod.
                      "\n", 2);
                next LOOP_PROD;
            } else {
                if ($::debug) {
                    ::log("reading status file\n", 2);
                }
                open(STATFILE, "< $statFile") or do {
                    ::log("ERROR reading $statFile: $!\n", 2);
                    ::log("!!! Evaluation of $prod stopped !!!\n\n", 2);
                    $ret = 1;
                    next LOOP_PROD;
                };
                binmode(STATFILE);
                $status = <STATFILE>;
                $status = uc(::chommp($status));
                close(STATFILE);
            }

            SWITCH: {
                if ($status =~ /^NO|^OK/io) {
                    # delete all old files from tempdir
                    ::deleteTempFiles($TempDir, $prod, 0);
                    ::log("status checked, skipping $prod\n", 2);
                    next LOOP_PROD;
                }
                if ($status =~ /^S(\d)E( .*)?$/i) {
                    my $scanNr = $1;
                    if (not $::reEval) {
                        # delete all old files from tempdir
                        ::deleteTempFiles($TempDir, $prod, 0);
                        # nothing to do
                        ::log("status checked, skipping $prod\n", 2);
                        next LOOP_PROD;
                    }
                    if ($::debug) {
                        ::log("status checked, re-evaluating $prod\n", 2);
                    }
                    # change status back from "evaluated" to "not evaluated"
                    $status = "S" . $scanNr;
                    # status changed => we fall through to the next case
                    ::updateStatus($statFile, $status) and do {
                        ::log("ERROR writing status file: $!\n", 2);
                        ::log("!!! Evaluation of $prod stopped !!!\n\n", 2);
                        $ret = 1;
                        next LOOP_PROD;
                    };
                }
                if ($status =~ /^S[1-3]( .*)?$/i) {
                    $status = substr($status, 0, 2);
                    # test if we really want to get the logfile from the
                    # archive
                    # this is only checked if we don't find any new logfiles
                    if (not $::skipReload) {
                        # delete all old files from tempdir
                        ::deleteTempFiles($TempDir, $prod, 0);
                        if ($::debug) {
                            ::log("status checked, loading logfile to ".
                                  $TempDir."\n", 2);
                        }
                        ::loadFile($ful, $LogDir, $TempDir) or do {
                            ::log("ERROR loading $ful: $!\n", 2);
                            ::log("!!! Evaluation of $prod stopped !!!\n\n",
                                  2);
                            $ret = 1;
                            next LOOP_PROD;
                        };
                        # look for postscans
                        my $fuls1 = $prod . $::FileExtScan1 . "." .
                        $::FileExtReport;
                        my $fuls2 = $prod . $::FileExtScan2 . "." .
                        $::FileExtReport;
                        my $fuls3 = $prod . $::FileExtScan3 . "." .
                        $::FileExtReport;
                        if ($status =~ /^S[2-3]$/io) {
                            # keep separate copy of first scan
                            copy($TempDir.$::sep.$ful,
                                 $TempDir.$::sep.$fuls1) or do {
                                ::log("ERROR creating $fuls1: $!\n", 2);
                                $ret = 1;
                            };
                            ::loadFile($fuls2, $LogDir, $TempDir) or do {
                                ::log("ERROR loading $fuls2: $!\n", 2);
                                ::log("!!! Evaluation of $prod stopped !!!".
                                      "\n\n", 2);
                                $ret = 1;
                                next LOOP_PROD;
                            };
                            ::appendFile($TempDir.$::sep.$ful,
                                         $TempDir.$::sep.$fuls2, 1) or do {
                                ::log("ERROR appending $fuls2 to $ful: $!\n",
                                      2);
                                ::log("!!! Evaluation of $prod stopped !!!".
                                      "\n\n", 2);
                                $ret = 1;
                                next LOOP_PROD;
                            };
                        }
                        if ($status =~ /^S3$/io) {
                            ::loadFile($fuls3, $LogDir, $TempDir) or do {
                                ::log("ERROR loading $fuls3: $!\n", 2);
                                ::log("!!! Evaluation of $prod stopped !!!".
                                      "\n\n", 2);
                                $ret = 1;
                                next LOOP_PROD;
                            };
                            ::appendFile($TempDir.$::sep.$ful,
                                         $TempDir.$::sep.$fuls3, 1) or do {
                                ::log("ERROR appending $fuls3 to $ful: $!\n",
                                      2);
                                ::log("!!! Evaluation of $prod stopped !!!".
                                      "\n\n", 2);
                                $ret = 1;
                                next LOOP_PROD;
                            };
                        }
                    } else {
                        # delete old files from tempdir but keep logfiles
                        ::deleteTempFiles($TempDir, $prod, 1);
                        if ($::debug) {
                            ::log("status checked, don't reload logfile\n", 2);
                        }
                    }
                    last SWITCH;
                }
                if ($status =~ /^-S1( .*)?$/i) {
                    $status = substr($status, 0, 3);
                    # delete all old files from tempdir
                    # this should not be needed here, but reset of
                    # product status doesn't care for that
                    ::deleteTempFiles($TempDir, $prod, 0);
                    # search for new logfile
                    if (not -e $NewLogDir.$::sep.$ful) {
                        ::log("no new logfiles, skipping $prod\n", 2);
                        next LOOP_PROD;
                    }
                    # the new logfile is copied to TempDir for evaluation,
                    # copied to LogDir as backup and removed from NewLogDir
                    # save a copy of the ful-file to LogDir
                    ::saveFile($ful, $NewLogDir, $LogDir) or do {
                        ::log("ERROR saving $ful: $!\n", 2);
                        ::log("!!! Evaluation of $prod stopped !!!\n\n", 2);
                        $ret = 1;
                        next LOOP_PROD;
                    };
                    move($NewLogDir.$::sep.$ful, $TempDir.$::sep.$ful) or do {
                        ::log("ERROR moving file: $!\n", 2);
                        ::log("!!! Evaluation of $prod stopped !!!\n\n", 2);
                        $ret = 1;
                        next LOOP_PROD;
                    };
                    # set new status
                    $status = "S1";
                    ::updateStatus($statFile, $status) and do {
                        ::log("ERROR writing status file: $!\n", 2);
                        ::log("!!! Evaluation of $prod stopped !!!\n\n", 2);
                        $ret = 1;
                        next LOOP_PROD;
                    };
                    last SWITCH;
                }
                if ($status =~ /^-S2( .*)?$/i) {
                    $status = substr($status, 0, 3);
                    my $fuls1 = $prod.$::FileExtScan1.".".$::FileExtReport;
                    my $fuls2 = $prod.$::FileExtScan2.".".$::FileExtReport;
                    my $txts2 = $prod.$::FileExtScan2.".".$::FileExtPSL;
                    # delete all old files from tempdir
                    ::deleteTempFiles($TempDir, $prod, 0);
                    if (not -e $NewLogDir.$::sep.$fuls2) {
                        ::log("no new Logfiles, skipping $prod\n", 2);
                        next LOOP_PROD;
                    }
                    ::saveFile($fuls2, $NewLogDir, $LogDir) or do {
                        ::log("ERROR saving $fuls2: $!\n", 2);
                        ::log("!!! Evaluation of $prod stopped !!!\n\n", 2);
                        $ret = 1;
                        next LOOP_PROD;
                    };
                    move($NewLogDir.$::sep.$fuls2, $TempDir.$::sep.$fuls2)
                    or do {
                        ::log("ERROR moving file: $!\n", 2);
                        ::log("!!! Evaluation of $prod stopped !!!\n\n", 2);
                        $ret = 1;
                        next LOOP_PROD;
                    };
                    copy($EvalDir.$::sep.$txts2, $TempDir.$::sep.$txts2)
                    or do {
                        ::log("ERROR copying file $txts2 from $EvalDir to ".
                              "$TempDir: $!\n", 2);
                        copy($NewLogDir.$::sep.$txts2, $TempDir.$::sep.$txts2)
                        or do {
                            ::log("ERROR copying file $txts2 from $NewLogDir".
                                  " to $TempDir: $!\n", 2);
                        $ret = 1;
                        };
                    };
                    unlink($NewLogDir.$::sep.$txts2) or do {
                        ::log("ERROR deleting file ".$NewLogDir.$::sep.$txts2.
                              ": $!\n", 2);
                    };
                    ::loadFile($ful, $LogDir, $TempDir) or do {
                        ::log("ERROR loading $ful: $!\n", 2);
                        ::log("!!! Evaluation of $prod stopped !!!\n\n", 2);
                        $ret = 1;
                        next LOOP_PROD;
                    };
                    copy($TempDir.$::sep.$ful, $TempDir.$::sep.$fuls1) or do {
                        ::log("ERROR creating $fuls1: $!\n", 2);
                        $ret = 1;
                    };
                    ::appendFile($TempDir.$::sep.$ful, $TempDir.$::sep.$fuls2, 1)
                    or do {
                        ::log("ERROR appending $fuls2 to $ful: $!\n", 2);
                        ::log("!!! Evaluation of $prod stopped !!!\n\n", 2);
                        $ret = 1;
                        next LOOP_PROD;
                    };
                    # set new status
                    $status = "S2";
                    ::updateStatus($statFile, $status) and do {
                        ::log("ERROR writing status file: $!\n", 2);
                        ::log("!!! Evaluation of $prod stopped !!!\n\n", 2);
                        $ret = 1;
                        next LOOP_PROD;
                    };
                    last SWITCH;
                }
                if ($status =~ /^-S3( .*)?$/i) {
                    $status = substr($status, 0, 3);
                    my $fuls1 = $prod.$::FileExtScan1.".".$::FileExtReport;
                    my $fuls2 = $prod.$::FileExtScan2.".".$::FileExtReport;
                    my $fuls3 = $prod.$::FileExtScan3.".".$::FileExtReport;
                    my $txts2 = $prod.$::FileExtScan2.".".$::FileExtPSL;
                    my $txts3 = $prod.$::FileExtScan3.".".$::FileExtPSL;
                    # delete all old files from tempdir
                    ::deleteTempFiles($TempDir, $prod, 0);
                    if (not -e $NewLogDir.$::sep.$fuls3) {
                        ::log("no new Logfiles, skipping $prod\n", 2);
                        next LOOP_PROD;
                    }
                    ::saveFile($fuls3, $NewLogDir, $LogDir) or do {
                        ::log("ERROR saving $fuls3: $!\n", 2);
                        ::log("!!! Evaluation of $prod stopped !!!\n\n", 2);
                        $ret = 1;
                        next LOOP_PROD;
                    };
                    move($NewLogDir.$::sep.$fuls3, $TempDir.$::sep.$fuls3)
                    or do {
                        ::log("ERROR moving file: $!\n", 2);
                        ::log("!!! Evaluation of $prod stopped !!!\n\n", 2);
                        $ret = 1;
                        next LOOP_PROD;
                    };
                    copy($EvalDir.$::sep.$txts3, $TempDir.$::sep.$txts3)
                    or do {
                        ::log("ERROR copying file $txts3 from $EvalDir to ".
                              "$TempDir: $!\n", 2);
                        copy($NewLogDir.$::sep.$txts3, $TempDir.$::sep.$txts3)
                        or do {
                            ::log("ERROR copying file $txts3 from $NewLogDir".
                                  " to $TempDir: $!\n", 2);
                            $ret = 1;
                        };
                    };
                    unlink($NewLogDir.$::sep.$txts3) or do {
                        ::log("ERROR deleting file ".$NewLogDir.$::sep.$txts3.
                              ": $!\n", 2);
                        $ret = 1;
                    };
                    ::loadFile($ful, $LogDir, $TempDir) or do {
                        ::log("ERROR loading $ful: $!\n", 2);
                        ::log("!!! Evaluation of $prod stopped !!!\n\n", 2);
                        $ret = 1;
                        next LOOP_PROD;
                    };
                    copy($TempDir.$::sep.$ful, $TempDir.$::sep.$fuls1) or do {
                        ::log("ERROR creating $fuls1: $!\n" ,2);
                        $ret = 1;
                    };
                    ::loadFile($fuls2, $LogDir, $TempDir) or do {
                        ::log("ERROR loading $fuls2: $!\n", 2);
                        ::log("!!! Evaluation of $prod stopped !!!\n\n", 2);
                        $ret = 1;
                        next LOOP_PROD;
                    };
                    copy($EvalDir.$::sep.$txts2, $TempDir.$::sep.$txts2)
                    or do {
                        ::log("ERROR copying file $txts2 from $EvalDir to ".
                              "$TempDir: $!\n", 2);
                        $ret = 1;
                    };
                    ::appendFile($TempDir.$::sep.$ful, $TempDir.$::sep.$fuls2, 1)
                    or do {
                        ::log("ERROR appending $fuls2 to $ful: $!\n", 2);
                        ::log("!!! Evaluation of $prod stopped !!!\n\n", 2);
                        $ret = 1;
                        next LOOP_PROD;
                    };
                    ::appendFile($TempDir.$::sep.$ful, $TempDir.$::sep.$fuls3, 1)
                    or do {
                        ::log("ERROR appending $fuls3 to $ful: $!\n", 2);
                        ::log("!!! Evaluation of $prod stopped !!!\n\n", 2);
                        $ret = 1;
                        next LOOP_PROD;
                    };
                    # set new status
                    $status = "S3";
                    ::updateStatus($statFile, $status) and do {
                        ::log("ERROR writing status file: $!\n", 2);
                        ::log("!!! Evaluation of $prod stopped !!!\n\n", 2);
                        $ret = 1;
                        next LOOP_PROD;
                    };
                    last SWITCH;
                }
                # catch all remaining possibilities
                ::log("ERROR: unknown status $status - skipping $prod\n", 2);
                $ret = 1;
                next LOOP_PROD;
            } # end of SWITCH
            # status is now one of S[1-3]

            # check if we really want to evaluate something
            if ($::archiveOnly) {
                if ($::debug) {
                    ::log("archiveOnly is set, skipping evaluation\n", 2);
                }
                next LOOP_PROD;
            }
            # now we have the up-to-date logfile in $TempDir

            ################## begin of logfile evaluation #################

            ::log("begin evaluation of $prod - $tb - $os\n", 2);

            # this uses the scanner key to transform the original logfile
            # into the normalized logfile (and some other files)
            if ($::debug) {
                ::log("generating normalized logfile (calling scanner key)".
                      "\n", 2);
            }
            normalizeLog($os, $tb, $prod, $ful, $TempDir)
            or do {
                # subroutine must give more detailed error description
                ::log("!!! Evaluation of $prod stopped !!!\n\n", 2);
                $ret = 1;
                next LOOP_PROD;
            };

            # now sort all lines and filter "excluded samples" to
            # get a valid normalized logfile
            if ($::debug) {
                ::log("generating normalized logfile (sorting, filtering)".
                      "\n", 2);
            }
            normalizeLogFinal($prod.".".$::FileExtNormUnsorted,
                              $prod.".".$::FileExtNorm,
                              $prod.".".$::FileExtExcludedFiles,
                              $excludeRef) or do {
                # subroutine must give more detailed error description
                ::log("!!! Evaluation of $prod stopped !!!\n\n", 2);
                $ret = 1;
                next LOOP_PROD;
            };


            # now there's a logfile in a normalized and sorted format
            # with all excluded samples filtered out

            # Special cases (= packed testbeds) are handled here
            if (exists $::lists{$tb}->{"packed"}) {
                if ($::debug) {
                    ::log("this is a packed testbed\n", 2);
                }
                # use this reference to access all archive filenames for
                # this testbed
                my $packRef = \%{$::lists{$tb}->{"packed"}->{"prg"}};

                # split logfile
                if ($::debug) {
                    ::log("splitting normalized logfile into ".
                          "packer-specific logfiles...\n", 2);
                }
                my @args = ($tb,
                            $prod . "." . $::FileExtNorm,
                            $prod . "." . $::FileExtPackedAdd,
                            $prod . "." . $::FileExtNormRemPacked,
                            $prod,
                            $::FileExtNorm,
                            $packRef);
                filterPacked(@args) or do {
                    # subroutine must give more detailed error description
                    ::log("!!! Evaluation of $prod stopped !!!\n\n", 2);
                    $ret = 1;
                    next LOOP_PROD;
                };

                # split scanned-but-ok-file
                # since some scanner-keys don't create this file, it is
                # created here (as empty file) to avoid errors when it's
                # missing
                my $file = $prod . "." . $::FileExtScannedOK;
                if (! -e $file) {
                    open(CREATE, "> $file") or do {
                        ::log("ERROR creating $file\n", 2);
                        $ret = 1;
                    };
                    close(CREATE);
                }
                if ($::debug) {
                    ::log("splitting scanned-but-Ok-file into ".
                          "packer-specific files...\n", 2);
                }
                @args = ($tb,
                         $prod . "." . $::FileExtScannedOK,
                         $prod . "." . $::FileExtPackedOKAdd,
                         $prod . "." . $::FileExtScanOKRemPacked,
                         $prod,
                         $::FileExtScannedOK,
                         $packRef);
                filterPackedOk(@args) or do {
                    # subroutine must give more detailed error description
                    ::log("!!! Evaluation of $prod stopped !!!\n\n", 2);
                    $ret = 1;
                    next LOOP_PROD;
                };

                # evaluate the logfile separately for each packer
                my $pack;
                foreach $pack (sort @tbPackList) {
                    if ($::debug) {
                        ::log("evaluating normalized logfile ".
                              "for program $pack...\n", 2);
                    }
                    my $prodPack = $prod . $pack;
                    my $checkPack = $prodPack . "." . $::FileExtControl;
                    my $isFP = 0;
                    my $anoRef = \%{$::lists{$tb}->{"evals"}->{$tb}->{"ano"}};
                    my @args_df = ($prodPack . "." . $::FileExtNorm,
                                   $prodPack . "." . $::FileExtDetFiles);
                    my @args_mis = ($::ListDir . $::sep . $::FilePreDirlist
                                    . $::listsAll{$tb}->{"id"}
                                    . $::FileExtArchiveFiles . "."
                                    . $::FileExtLists,
                                    $prodPack . "." . $::FileExtDetFiles,
                                    $prodPack . "." . $::FileExtMissedFiles,
                                    $prodPack . "." . $::FileExtAnoFiles,
                                    $prodPack . "." . $::FileExtGarbageFiles,
                                    $anoRef);

                    if ($::debug > 1) {
                        ::log("anoList for $tb:\n", 3);
                        foreach (keys %{$anoRef}) { ::log($_."\n", 4); }
                    }
                    # split logfile for FP testbeds
                    if (exists $::lists{$tb}->{"fp"}) {
                        if ($::debug) {
                            ::log("split logfile for FP testbeds...\n", 2);
                        }
                        $isFP = 1;
                        my $fpList = $::lists{$tb}->{"fp"};
                        substr($fpList, 2, 1) = $::FileIdFP;
                        my $prodFP = $prodPack . "_" . $::FileIdFP;
                        filterLogFP($prodPack, $::lists{$tb}->{"main"}) or do {
                            ::log("!!! Evaluation of $prod stopped !!!\n\n",
                                  2);
                            $ret = 1;
                            next LOOP_PROD;
                        };
                        # some evaluation steps are done twice: for fp's and for
                        # normal samples
                        @args_df = ($prodFP . "." . $::FileExtEval,
                                    $prodPack . "." . $::FileExtDetFPFiles);
                        # create a list of ALL detected files in this testbed
                        my $retVal = eval_df(@args_df);
                        if ($retVal == -1) {
                            $::data{"DF_Filtered"}->{$os}->{$tb}->{$prodFP} =
                                0;
                            ::log("ERROR creating 'detected files' list for ".
                                  "false positives\n", 2);
                            ::log("!!! Evaluation of $prod stopped !!!\n\n",
                                  2);
                            $ret = 1;
                            next LOOP_PROD;
                        } else {
                            $::data{"DF_Filtered"}->{$os}->{$tb}->{$prodFP} =
                                $retVal;
                        }

                        @args_mis = ($::ListDir . $::sep . $fpList . "."
                                     . $::FileExtSamples,
                                     $prod . "." . $::FileExtDetFPFiles,
                                     $prod . "." . $::FileExtMissedFPFiles,
                                     $prodFP . "." . $::FileExtAnoFiles,
                                     $prodFP . "." . $::FileExtGarbageFiles,
                                     $anoRef);
                        eval_mis(@args_mis) or do {
                            ::log("ERROR creating 'missed samples' list for ".
                                  "false positives\n", 2);
                            ::log("!!! Evaluation of $prod stopped !!!\n\n",
                                  2);
                            $ret = 1;
                            next LOOP_PROD;
                        };
                        # we need to use different input file for fp-testbeds to
                        # get a clear separation between fp's and samples
                        @args_df = ($prodPack . "." . $::FileExtEval,
                                    $prodPack . "." . $::FileExtDetFiles);
                        @args_mis = ($::ListDir . $::sep . $tb . "."
                                     . $::FileExtSamples,
                                     $prod . "." . $::FileExtDetFiles,
                                     $prod . "." . $::FileExtMissedFiles,
                                     $prod . "." . $::FileExtAnoFiles,
                                     $prod . "." . $::FileExtGarbageFiles,
                                     $anoRef);
                    } else {
                        if ($::debug) {
                            ::log("creating $prod.$::FileExtEval\n", 2);
                        }
                        # No need to filter anything with no samples to exclude
                        # Just copy the whole logfile
                        copy($prodPack . "." . $::FileExtNorm,
                             $prodPack . "." . $::FileExtEval) or do {
                            ::log("ERROR: copy ".$prodPack,".".$::FileExtNorm.
                                  " to ".$prodPack.".".$::FileExtEval.
                                  " failed\n", 2);
                            ::log("!!! Evaluation of $prod stopped !!!\n\n",
                                  2);
                            $ret = 1;
                            next LOOP_PROD;
                        };
                    }

                    # create a list of ALL detected files in this archive
                    my $retVal = eval_df(@args_df);
                    if ($retVal == -1) {
                        $::data{"DF_Filtered"}->{$os}->{$tb}->{$prodPack} = 0;
                        # subroutine must give more detailed error description
                        ::log("ERROR creating 'detected files' list\n", 2);
                        ::log("!!! Evaluation of $prod stopped !!!\n\n", 2);
                        $ret = 1;
                        next LOOP_PROD;
                    } else {
                        $::data{"DF_Filtered"}->{$os}->{$tb}->{$prodPack} =
                            $retVal;
                    }

                    eval_mis(@args_mis) or do {
                        # subroutine must give more detailed error description
                        ::log("ERROR creating missed samples\n", 2);
                        ::log("!!! Evaluation of $prod stopped !!!\n\n", 2);
                        $ret = 1;
                        next LOOP_PROD;
                    };

                    # trim sbo-list
                    $retVal = eval_sbo($prodPack .".". $::FileExtScannedOK,
                                          $prodPack .".". $::FileExtOKFiles,
                                          $excludeRef);
                    if ($retVal == -1) {
                        $::data{"OF_Filtered"}->{$os}->{$tb}->{$prodPack} = 0;
                        ::log("ERROR evaluating ".$prodPack.".".
                              $::FileExtScannedOK."\n", 2);
                        ::log("!!! Evaluation of $prod stopped !!!\n\n", 2);
                        $ret = 1;
                        next LOOP_PROD;
                    } else {
                        $::data{"OF_Filtered"}->{$os}->{$tb}->{$prodPack} =
                            $retVal;
                    }

                    # create postscan list (empty if no postscan required)
                    # and some other files (for error checking)
                    @args = ($prodPack,
                             $::ListDir . $::sep . $::FilePreDirlist
                             . $::listsAll{$tb}->{"id"}
                             . $::FileExtArchiveFiles . "."
                             . $::FileExtLists,
                             $isFP);
                    eval_postscan(@args) or do {
                        # subroutine must give more detailed error description
                        ::log("ERROR creating postscan list\n", 2);
                        $ret = 1;
                    };

                    # check testbed evaluation for this packer
                    checkTbCommon($os, $tb, $prodPack, $checkPack,
                                  $::ListDir . $::sep . $::FilePreDirlist .
                                  $::lists{$tb}->{"id"} .
                                  $::FileExtArchiveFiles .
                                  "." . $::FileExtLists);

                    # The remaining evaluation steps must be repeated for each
                    # evaluation list and are therefore done in a subroutine
                    # (this includes data checks).
                    doLists($os, $tb, $prod, $checkPack, $prodPack) or do {
                        # subroutine must give more detailed error description
                        ::log("!!! Evaluation of $prod stopped !!!\n\n", 2);
                        $ret = 1;
                        next LOOP_PROD;
                    };
                }

                # evaluate the archives themselves
                if ($::debug) {
                    ::log("evaluating normalized logfile for archives...\n",
                          2);
                }
                $pack = $::FileExtArchiveFiles;
                my $prodPack = $prod . $pack;
                my $checkPack = $prodPack . "." . $::FileExtControl;
                my $isFP = 0;
                my $anoRef = \%{$::lists{$tb}->{"packed"}->{"ano"}};
                my @args_df = ($prodPack . "." . $::FileExtNorm,
                               $prodPack . "." . $::FileExtDetFiles);
                my @args_mis = ($::ListDir . $::sep . $::FilePreDirlist
                                . $::listsAll{$tb}->{"id"} . "."
                                . $::FileExtLists,
                                $prodPack . "." . $::FileExtDetFiles,
                                $prodPack . "." . $::FileExtMissedFiles,
                                $prodPack . "." . $::FileExtAnoFiles,
                                $prodPack . "." . $::FileExtGarbageFiles,
                                $anoRef);

                # split logfile for FP testbeds
                if (exists $::lists{$tb}->{"fp"}) {
                    if ($::debug) {
                        ::log("split logfile for FP testbeds...\n", 2);
                    }
                    $isFP = 1;
                    my $fpList = $::lists{$tb}->{"fp"};
                    substr($fpList, 2, 1) = $::FileIdFP;
                    my $prodFP = $prodPack . "_" . $::FileIdFP;
                    filterLogFP($prodPack, $::lists{$tb}->{"main"}) or do {
                        ::log("!!! Evaluation of $prod stopped !!!\n\n", 2);
                        $ret = 1;
                        next LOOP_PROD;
                    };
                    # some evaluation steps are done twice: for fp's and for
                    # normal samples
                    @args_df = ($prodFP . "." . $::FileExtEval,
                                $prodPack . "." . $::FileExtDetFPFiles);
                    # create a list of ALL detected files in this testbed
                    my $retVal = eval_df(@args_df);
                    if ($retVal == -1) {
                        $::data{"DF_Filtered"}->{$os}->{$tb}->{$prodFP} = 0;
                        ::log("ERROR creating 'detected files' list for ".
                              "false positives\n", 2);
                        ::log("!!! Evaluation of $prod stopped !!!\n\n", 2);
                        $ret = 1;
                        next LOOP_PROD;
                    } else {
                        $::data{"DF_Filtered"}->{$os}->{$tb}->{$prodFP} =
                            $retVal;
                    }

                    @args_mis = ($::ListDir . $::sep . $fpList . "."
                                 . $::FileExtSamples,
                                 $prod . "." . $::FileExtDetFPFiles,
                                 $prod . "." . $::FileExtMissedFPFiles,
                                 $prodFP . "." . $::FileExtAnoFiles,
                                 $prodFP . "." . $::FileExtGarbageFiles,
                                 $anoRef);
                    eval_mis(@args_mis) or do {
                        ::log("ERROR creating 'missed samples' list for ".
                              "false positives\n", 2);
                        ::log("!!! Evaluation of $prod stopped !!!\n\n", 2);
                        $ret = 1;
                        next LOOP_PROD;
                    };
                    # we need to use different input file for fp-testbeds to
                    # get a clear separation between fp's and samples
                    @args_df = ($prodPack . "." . $::FileExtEval,
                                $prodPack . "." . $::FileExtDetFiles);
                    @args_mis = ($::ListDir . $::sep . $tb . "."
                                 . $::FileExtSamples,
                                 $prod . "." . $::FileExtDetFiles,
                                 $prod . "." . $::FileExtMissedFiles,
                                 $prod . "." . $::FileExtAnoFiles,
                                 $prod . "." . $::FileExtGarbageFiles,
                                 $anoRef);
                } else {
                        # No need to filter anything with no samples to exclude
                        # Just copy the whole logfile
                        copy($prodPack . "." . $::FileExtNorm,
                             $prodPack . "." . $::FileExtEval) or do {
                            ::log("ERROR: copy ".$prodPack.".".$::FileExtNorm.
                                  " to ".$prodPack.".".$::FileExtEval.
                                  " failed\n", 2);
                            ::log("!!! Evaluation of $prod stopped !!!\n\n",
                                  2);
                            $ret = 1;
                            next LOOP_PROD;
                        };
                }

                # create a list of ALL detected files in this archive
                my $retVal = eval_df(@args_df);
                if ($retVal == -1) {
                    $::data{"DF_Filtered"}->{$os}->{$tb}->{$prodPack} = 0;
                    # subroutine must give more detailed error description
                    ::log("ERROR creating 'detected files' list\n", 2);
                    ::log("!!! Evaluation of $prod stopped !!!\n\n", 2);
                    $ret = 1;
                    next LOOP_PROD;
                } else {
                    $::data{"DF_Filtered"}->{$os}->{$tb}->{$prodPack} =
                        $retVal;
                }

                eval_mis(@args_mis) or do {
                    # subroutine must give more detailed error description
                    ::log("ERROR creating missed samples\n", 2);
                    ::log("!!! Evaluation of $prod stopped !!!\n\n", 2);
                    $ret = 1;
                    next LOOP_PROD;
                };

                # trim sbo-list
                $retVal = eval_sbo($prodPack .".". $::FileExtScannedOK,
                                  $prodPack .".". $::FileExtOKFiles,
                                  $excludeRef);
                if ($retVal == -1) {
                    $::data{"OF_Filtered"}->{$os}->{$tb}->{$prodPack} = 0;
                    ::log("ERROR evaluating ".$prodPack.".".
                          $::FileExtScannedOK."\n", 2);
                    ::log("!!! Evaluation of $prod stopped !!!\n\n", 2);
                    $ret = 1;
                    next LOOP_PROD;
                } else {
                    $::data{"OF_Filtered"}->{$os}->{$tb}->{$prodPack} =
                        $retVal;
                }

                # create postscan list (empty if no postscan required)
                # and some other files (for error checking)
                @args = ($prodPack,
                         $::ListDir . $::sep . $::FilePreDirlist
                         . $::listsAll{$tb}->{"id"} . "." . $::FileExtLists,
                         $isFP);
                eval_postscan(@args) or do {
                    # subroutine must give more detailed error description
                    ::log("ERROR creating postscan list\n", 2);
                    $ret = 1;
                };

                # check testbed evaluation for archive evaluation
                checkTbCommon($os, $tb, $prodPack, $checkPack,
                              $::ListDir . $::sep . $::FilePreDirlist .
                              $::lists{$tb}->{"id"} ."." . $::FileExtLists);
                doArchiveList($os, $tb, $prod, $checkPack, $prodPack) or do {
                    # subroutine must give more detailed error description
                    ::log("!!! Evaluation of $prod stopped !!!\n\n", 2);
                    $ret = 1;
                    next LOOP_PROD;
                };

                # and now put these together to create some testbed data
                eval_df_Packed($prod) or do {
                    ::log("ERROR creating 'detected files' list\n", 2);
                    ::log("!!! Evaluation of $prod stopped !!!\n\n", 2);
                    $ret = 1;
                    next LOOP_PROD;
                };

                eval_mis_Packed($prod, $tb, $packRef) or do {
                    # subroutine must give more detailed error description
                    ::log("ERROR creating missed samples\n", 2);
                    ::log("!!! Evaluation of $prod stopped !!!\n\n", 2);
                    $ret = 1;
                    next LOOP_PROD;
                };

                eval_postscan_Packed($prod, $tb, $packRef) or do {
                    ::log("ERROR creating postscan list\n", 2);
                    ::log("!!! Evaluation of $prod stopped !!!\n\n", 2);
                    $ret = 1;
                    next LOOP_PROD;
                };

                # do some additional checks for packed testbeds
                # and create a single control file
                if ($::debug) {
                    ::log("checking testbed data...\n", 2);
                }
                checkPackedAll($os, $tb, $prod, $checkFile);


            } else {
            # handle all other testbeds than 'packed'
                if ($::debug) {
                    ::log("evaluating normalized logfile...\n", 2);
                }
                my $isFP = 0;
                my $anoRef = \%{$::lists{$tb}->{"evals"}->{$tb}->{"ano"}};
                my @args_df = ($prod . "." . $::FileExtNorm,
                               $prod . "." . $::FileExtDetFiles);
                my @args_mis = ($::ListDir . $::sep . $::FilePreDirlist .
                                $::listsAll{$tb}->{"id"}.".".$::FileExtLists,
                                $prod . "." . $::FileExtDetFiles,
                                $prod . "." . $::FileExtMissedFiles,
                                $prod . "." . $::FileExtAnoFiles,
                                $prod . "." . $::FileExtGarbageFiles,
                                $anoRef);
                # split logfile for FP testbeds
                if (exists $::lists{$tb}->{"fp"}) {
                    if ($::debug) {
                        ::log("split logfile for FP testbeds...\n", 2);
                    }
                    $isFP = 1;
                    my $fpList = $::lists{$tb}->{"fp"};
                    my $prodFP = $prod . "_" . $::FileIdFP;
                    filterLogFP($prod, $::listsAll{$tb}->{"main"}) or do {
                        # subroutine must give more detailed error description
                        ::log("!!! Evaluation of $prod stopped !!!\n\n", 2);
                        $ret = 1;
                        next LOOP_PROD;
                    };
                    # some evaluation steps are done twice: for fp's and for
                    # normal samples
                    @args_df = ($prodFP . "." . $::FileExtEval,
                                $prod . "." . $::FileExtDetFPFiles);
                    # create a list of ALL detected files in this testbed
                    my $retVal = eval_df(@args_df);
                    if ($retVal == -1) {
                        $::data{"DF_Filtered"}->{$os}->{$tb}->{$prodFP} = 0;
                        ::log("ERROR creating 'detected files' list for ".
                              "false positives\n", 2);
                        ::log("!!! Evaluation of $prod stopped !!!\n\n", 2);
                        $ret = 1;
                        next LOOP_PROD;
                    } else {
                        $::data{"DF_Filtered"}->{$os}->{$tb}->{$prodFP} =
                            $retVal;
                    }

                    @args_mis = ($::ListDir.$::sep.$fpList.".".$::FileExtSamples,
                                    $prod . "." . $::FileExtDetFPFiles,
                                    $prod . "." . $::FileExtMissedFPFiles,
                                    $prodFP . "." . $::FileExtAnoFiles,
                                    $prodFP . "." . $::FileExtGarbageFiles,
                                    $anoRef);
                    eval_mis(@args_mis) or do {
                        ::log("ERROR creating 'missed samples' list for ".
                              "false positives\n", 2);
                        ::log("!!! Evaluation of $prod stopped !!!\n\n", 2);
                        $ret = 1;
                        next LOOP_PROD;
                    };
                    # we need to use different input file for fp-testbeds to
                    # get a clear separation between fp's and samples
                    @args_df = ($prod . "." . $::FileExtEval,
                                $prod . "." . $::FileExtDetFiles);
                    @args_mis = ($::ListDir .$::sep. $tb .".". $::FileExtSamples,
                                    $prod . "." . $::FileExtDetFiles,
                                    $prod . "." . $::FileExtMissedFiles,
                                    $prod . "." . $::FileExtAnoFiles,
                                    $prod . "." . $::FileExtGarbageFiles,
                                    $anoRef);
                } elsif (exists $::lists{$tb}->{"single"}) {
                    # there is only one evaluationlist
                    if ($::debug) {
                        ::log("creating $prod.$::FileExtEval\n", 2);
                    }
                        # No need to filter anything with no samples to exclude
                        # Just copy the whole logfile
                        copy($prod . "." . $::FileExtNorm,
                             $prod . "." . $::FileExtEval) or do {
                            ::log("ERROR: copy ".$prod.".".$::FileExtNorm.
                                  " to ".$prod.".".$::FileExtEval.
                                  " failed\n", 2);
                            ::log("!!! Evaluation of $prod stopped !!!\n\n",
                                  2);
                            $ret = 1;
                            next LOOP_PROD;
                        };
                } else {
                    ::log("ERROR: can't handle multiple evalution lists for ".
                          "this testbed.\n", 2);
                    ::log("!!! Evaluation of $prod stopped !!!\n\n", 2);
                    $ret = 1;
                    next LOOP_PROD;
### This MUST be implemented when there is more than one evallist (besides fp lists) !!!
#                    filterLog($prod, $::listsAll{$tb}->{"main"}) or do {
#                        # subroutine must give more detailed error description
#                        ::log("!!! Evaluation of $prod stopped !!!\n\n", 2);
#                        next LOOP_PROD;
#                    };
###
                }

                # create a list of ALL detected files in this testbed
                my $retVal = eval_df(@args_df);
                if ($retVal == -1) {
                    $::data{"DF_Filtered"}->{$os}->{$tb}->{$prod} = 0;
                    ::log("ERROR creating 'detected files' list\n", 2);
                    ::log("!!! Evaluation of $prod stopped !!!\n\n", 2);
                    $ret = 1;
                    next LOOP_PROD;
                } else {
                    $::data{"DF_Filtered"}->{$os}->{$tb}->{$prod} =
                        $retVal;
                }

                eval_mis(@args_mis) or do {
                    # subroutine must give more detailed error description
                    ::log("ERROR creating 'missed samples' list\n", 2);
                    ::log("!!! Evaluation of $prod stopped !!!\n\n", 2);
                    $ret = 1;
                    next LOOP_PROD;
                };

                # since some scanner-keys don't create this file, it is
                # created here (as empty file) to avoid errors when it's
                # missing
                my $file = $prod . "." . $::FileExtScannedOK;
                if (! -e $file) {
                    open(CREATE, "> $file") or do {
                        ::log("ERROR creating $file\n", 2);
                    };
                    $ret = 1;
                    close(CREATE);
                }
                # trim sbo-list
                $retVal = eval_sbo($prod . "." . $::FileExtScannedOK,
                                   $prod . "." . $::FileExtOKFiles,
                                   $excludeRef);
                if ($retVal == -1) {
                    $::data{"OF_Filtered"}->{$os}->{$tb}->{$prod} = 0;
                    ::log("ERROR evaluating $prod.$::FileExtScannedOK\n", 2);
                    ::log("!!! Evaluation of $prod stopped !!!\n\n", 2);
                    $ret = 1;
                    next LOOP_PROD;
                } else {
                    $::data{"OF_Filtered"}->{$os}->{$tb}->{$prod} =
                        $retVal;
                }

                # create postscan list (empty if no postscan required)
                # and some other files (for error checking)
                my @args_ps = ($prod,
                               $::ListDir . $::sep . $::FilePreDirlist
                               . $::listsAll{$tb}->{"id"} . "."
                               . $::FileExtLists,
                               $isFP);
                eval_postscan(@args_ps) or do {
                    # subroutine must give more detailed error description
                    ::log("ERROR creating postscan list\n", 2);
                    $ret = 1;
                };

                # check testbed evaluation
                if ($::debug) {
                    ::log("checking testbed data...\n", 2);
                }
                checkTb($os, $tb, $prod, $checkFile,
                        $::ListDir . $::sep . $::FilePreDirlist .
                        $::lists{$tb}->{"id"} . "." . $::FileExtLists);
                checkTbCommon($os, $tb, $prod, $checkFile,
                              $::ListDir . $::sep . $::FilePreDirlist .
                              $::lists{$tb}->{"id"} . "." . $::FileExtLists);

                doLists($os, $tb, $prod, $checkFile, $prod) or do {
                    # subroutine must give more detailed error description
                    ::log("!!! Evaluation of $prod stopped !!!\n\n", 2);
                    $ret = 1;
                    next LOOP_PROD;
                };
            }

            ############# evaluation is done - now save the data ###########
            # rename postscan list and make control file backup
            my $pslFile = $prod . "." . $::FileExtPostscan;
            my $conFile = $prod . "." . $::FileExtControl;
            my ($bakFile, $txtFile);
            if ($status =~ /^S1/i) {
                $bakFile = $prod . $::FileExtScan1 . "." . $::FileExtControl;
                $txtFile = $prod . $::FileExtScan2 . "." . $::FileExtPSL;
                if (-s $pslFile) {
                    copy($pslFile, $txtFile) or do {
                        ::log("ERROR renaming postscan list: $!\n", 2);
                        $ret = 1;
                    };
                }
            } elsif ($status =~ /^S2/i) {
                $bakFile = $prod . $::FileExtScan2 . "." . $::FileExtControl;
                $txtFile = $prod . $::FileExtScan3 . "." . $::FileExtPSL;
                if (-s $pslFile) {
                    copy($pslFile, $txtFile) or do {
                        ::log("ERROR renaming postscan list: $!\n", 2);
                        $ret = 1;
                    };
                }
            } elsif ($status =~ /^S3/i) {
                $bakFile = $prod . $::FileExtScan3 . "." . $::FileExtControl;
                # don't create postscan-textfile after last postscan
            }
            # always create backup copy of the control file (that is not
            # overwritten during postscan evaluation)
            copy($conFile, $bakFile) or do {
                ::log("ERROR making control file backup: $!\n", 2);
                $ret = 1;
            };

            ###### save results (TempDir is no safe place for data) ########
            # save important files as plain text additionally to...
            if ($::debug) {
                ::log("saving results in EvalDir\n", 2);
            }
            # datFiles are saved earlier, all other files are copied here
            my @files =  ($prod . "." . $::FileExtControl,
                          $prod . "." . $::FileExtDetFiles,
                          $prod . "." . $::FileExtMissedFiles,
                          $prod . "." . $::FileExtPostscan);

            if (exists $::lists{$tb}->{"fp"}) {
                push @files, $prod .".". $::FileExtDetFPFiles;
                push @files, $prod .".". $::FileExtMissedFPFiles;
            }
            my $tempFile = $prod . $::FileExtScan2 . "." . $::FileExtPSL;
            if (-e $tempFile) { push @files, $tempFile; }

            $tempFile = $prod . $::FileExtScan3 . "." . $::FileExtPSL;
            if (-e $tempFile) { push @files, $tempFile; }

#            $tempFile = $prod . $::FileExtScan1 . "." . $::FileExtControl;
#            if (-e $tempFile) { push @files, $tempFile; }

#            $tempFile = $prod . $::FileExtScan2 . "." . $::FileExtControl;
#            if (-e $tempFile) { push @files, $tempFile; }

#            $tempFile = $prod . $::FileExtScan3 . "." . $::FileExtControl;
#            if (-e $tempFile) { push @files, $tempFile; }

            foreach $file (@files) {
                copy($TempDir.$::sep.$file, $EvalDir.$::sep.$file) or do {
                    ::log("ERROR saving data in $EvalDir: $!\n", 2);
                    ::log("      Evaluation of $prod is done.\n", 2);
                    ::log("      Copy data from $TempDir by hand\n\n", 2);
                    $ret = 1;
                };
            }
            # ...saving all files in one compressed file
            ::archiveFiles($prod, $TempDir, $EvalDir, @files) or do {
                ::log("ERROR saving data in $EvalDir: $!\n", 2);
                ::log("      Evaluation of $prod is done.\n", 2);
                ::log("      Copy data from $TempDir by hand (and update ".
                      "status)\n", 2);
                ::log("      to avoid re-evaluation.\n\n", 2);
                $ret = 1;
                next LOOP_PROD;
            };

            # evaluation done - set status to "evaluated"
            $status = $status . "E";
            ::updateStatus($statFile, $status) and do {
                ::log("ERROR writing status file: $!\n", 2);
                ::log("      Evaluation of $prod is done.\n", 2);
                ::log("      Update ".$prod.$::FileExtStatus.
                      "by hand to avoid re-evaluation.\n\n", 2);
                $ret = 1;
            };
            ::log("evaluation of $prod - $tb - $os done\n", 2);
        }  #end of foreach prod LOOP_PROD

    }  #end of foreach tb
    }  #end of foreach os

    return $ret;
} # end of subroutine




######################### evaluation subroutines ###########################

# this uses the scanner key to transform the original logfile into the
# normalized logfile (and some other files)
sub normalizeLog {
    my $os = $_[0];
    my $tb = $_[1];
    my $prod = $_[2];
    my $ful = $_[3];
    my $TempDir = $_[4];
    # These are the product-specific scanner-keys (formerly known as
    # scanner-patterns)
    # Although only one scanner-key is needed (and used) there can be
    # more than one to choose from:
    # All new scanner-keys are implemented as perl scripts.
    # !!! Old awk scripts can't be used anymore !!!
    # To avoid having to keep in sync many files, there is
    # one general scanner key for all operating systems (prod.pl)
    # and there might be OS-specific scanner-keys (prod_os.pl).
    # The one that is actually used is selected with this priority:
    #   1. OS-specific perl script (prod_os.pl)
    #   2. general perl script (prod.pl)
    # Both uppercase and lowercase filenames are probed to avoid errors
    # on case-sensitive operating systems

    # The scanner key is called with an absolute path ($::KeyDir is already
    # in correct case and syntax for the current system)
    my $key = $::KeyDir . $::sep;
    # find the right key file
    my $keyTryOs = uc($prod . "_" . $os . ".PL");
    my $keyTryOs2 = lc($prod . "_" . $os . ".pl");
    my $keyTry = uc($prod . ".PL");
    my $keyTry2 = lc($prod . ".pl");
    if (-e $key.$keyTryOs) {
        $key .= $keyTryOs;
    } elsif (-e $key.$keyTryOs2) {
        $key .= $keyTryOs2;
    } elsif (-e $key.$keyTry) {
        $key .= $keyTry;
    } elsif (-e $key.$keyTry2) {
        $key .= $keyTry2;
    } else {
        ::log("ERROR: no scanner key found\n", 3);
        return 0;
    }

    # this is the interface to hand filenames and other data to the scanner key
    %::KeyAPI = ();
    $::KeyAPI{"constEOL"} = $::EOL;
    $::KeyAPI{"constUNIX"} = $::OS_UNIX;
    $::KeyAPI{"constWINDOWS"} = $::OS_WINDOWS;
    $::KeyAPI{"debug"} = $::debug;
    $::KeyAPI{"logRef"} = \&::log;
    $::KeyAPI{"DebugFile"} = $::DebugFile;
    $::KeyAPI{"nameUI"} = $::VirusNameUI;
    $::KeyAPI{"inLog"} = $ful;
    $::KeyAPI{"outNorm"} = $prod . "." . $::FileExtNormUnsorted;
    $::KeyAPI{"outOkay"} = $prod . "." . $::FileExtScannedOK;
    $::KeyAPI{"outEval"} = $prod . "." . $::FileExtAuswert;
    $::KeyAPI{"outEvalRedundant"} = $prod . "." . $::FileExtAuswertRedundant;
    $::KeyAPI{"outRest"} = $prod . "." . $::FileExtReportRem;
    $::KeyAPI{"outCut"} = $prod . "." . $::FileExtCutList;
    $::KeyAPI{"outControl"} = $prod . "." . $::FileExtControl;
    $::KeyAPI{"outCorrupt"} = $prod . "." . $::FileExtCorrupt;
    $::KeyAPI{"os"} = $os;
    $::KeyAPI{"testbed"} = $tb;
    $::KeyAPI{"product"} = $prod;
    $::KeyAPI{"testbedID"} = $::lists{$tb}->{"id"};
    $::KeyAPI{"syntax"} = $::configSyntax;
    $::KeyAPI{"sep"} = $::rps;
    $::KeyAPI{"sepRun"} = $::sep;
    $::KeyAPI{"pathUnix"} = $::lists{$tb}->{"path"}->{$::OS_UNIX};
    $::KeyAPI{"pathWindows"} = $::lists{$tb}->{"path"}->{$::OS_WINDOWS};
    if ($::configSyntax eq $::OS_WINDOWS) {
        $::KeyAPI{"testbedRoot"} = $::lists{$tb}->{"path"}->{$::OS_WINDOWS};
    } else {
        $::KeyAPI{"testbedRoot"} = $::lists{$tb}->{"path"}->{$::OS_UNIX};
    }
    $::KeyAPI{"tempDir"} = $::TempBase;
    $::KeyAPI{"binDir"} = $::ScriptDir;
    # data for packed testbeds
    $::KeyAPI{"packed"} = "";
    $::KeyAPI{"packRef"} = "";
    # the regular expression for sample names is different for packed testbeds
    # (it uses the driveletter of the unpacked reference testbed)
    if (exists $::lists{$tb}->{"packed"}) {
        $::KeyAPI{"packed"} = $::lists{$tb}->{"packed"}->{"type"};
        $::KeyAPI{"packRef"} = \%{$::lists{$tb}->{"packed"}->{"prg"}};
        my $refTb = $tb;
        substr($refTb, 3, 2) = "__";
        $::KeyAPI{"samplenameRE"} = $::listsAll{$refTb}->{"id"} . $::samplenameRE;
    } else {
        $::KeyAPI{"samplenameRE"} = $::lists{$tb}->{"id"} . $::samplenameRE;
    }

    # data for boot testbeds
    $::KeyAPI{"boot"} = 0;
    if (substr($tb, 0, 1) eq "B") {
        $::KeyAPI{"boot"} = 1;
    }

    if ($::debug > 2) {
        ::log("KeyAPI data dump before calling scanner-key\n", 4);
        foreach my $k (keys %::KeyAPI) {
            ::log("key='$k'  -  value='$::KeyAPI{$k}'\n", 5);
        }
    }
    ::log("calling $key\n", 3);

    # execute scanner key script
    my $ok = do $key;
    if (not defined($ok)) {
        $ok = 0;
        if ($@) {
            ::log("ERROR: perl couldn't compile scanner-key '$key':\n", 3);
            ::log("       $@\n", 3);
        } else {
            ::log("ERROR: perl couldn't read scanner-key '$key'\n", 3);
        }
    } elsif (not $ok) {
        ::log("ERROR in scanner key $key\n", 3);
    } else {
        ::log("scanner key returned success\n", 3);
    }
    # since the API is globally visible and its contents are persistent
    # throughout the whole evaluation it is better to clear it completely
    # after each use
    undef %::KeyAPI;

    return $ok;
}


# this subroutine iterates through all evaluation lists and does all
# remaining evaluation and checking steps
sub doLists {
    # some data that must be passed by the caller
    my $os = $_[0];        # name of current operating system
    my $tb = $_[1];        # name of current testbed
    my $prod = $_[2];      # this is the 3-char code of the current product
    my $checkFile = $_[3]; # name of the file for check data
    my $prodPart = $_[4];  # this is the string that is used as filename
                           # prefix for this evaluation. This is mostly
                           # equal to $prod but can also be different,
                           # eg. $prod="FPR" but $ProdPart="FPRZIP"
    my $EvalDir = $::EvalBase . $::sep . $os . $::sep . $tb;
    my $TempDir = $::TempBase . $::sep . $os . $::sep . $tb;

    my $evalList;
    LOOP_EL: foreach $evalList (keys %{$::lists{$tb}->{"evals"}}) {
        if ($::debug) {
            ::log("starting evaluation for list $evalList ...\n", 3);
        }
        # get the ID character for this evaluation list
        # all files created here are marked with this ID
        my $id = $::lists{$tb}->{"evals"}->{$evalList}->{"id"};
        my $prodEval = $prodPart;
        # the "main evaluation list" for the whole testbed is
        # recognized by an empty ID ("") and should have no
        # additional character in the filenames
        if ($id ne "") {
            $prodEval .= "_" . $id;
        }

        # process normalized logfile - file level evaluation
        if ($::debug) {
            ::log("doing file level evaluation\n", 3);
        }
        my @files = ($prodEval . "." . $::FileExtEval,
                     $::ListDir . $::sep . $evalList,
                     $prodEval . "." . $::FileExtDetVir,
                     $prodEval . "." . $::FileExtUnrId,
                     $prodEval . "." . $::FileExtMultInf,
                     $prodEval . "." . $::FileExtMultScan,
                     $prodEval . "." . $::FileExtFilteredAno);
        my $anoRef = \%{$::lists{$tb}->{"evals"}->{$evalList}->{"ano"}};
        eval_files(@files, $anoRef) or do {
            # more error details should be given in the subroutine
            ::log("ERROR: evaluating files failed\n", 3);
            return 0;
        };

        # the last step is just counting of the results from the
        # previous step to get the total sums for viruses and samples
        if ($::debug) {
            ::log("counting samples\n", 3);
        }
        my @args = ($prodEval . "." . $::FileExtDetVir,
                    $prodEval . "." . $::FileExtResult,
                    $::lists{$tb}->{"evals"}->{$evalList}->{"sumV"},
                    $::lists{$tb}->{"evals"}->{$evalList}->{"sumF"},
                    $prod,
                    $id);
        eval_variants(@args) or do {
            # more error details should be given in the subroutine
            return 0;
        };

        if ($::debug) {
            ::log("checking results\n", 3);
        }

        @args = ($tb, $evalList, $prodEval, $checkFile,
         $::lists{$tb}->{"evals"}->{$evalList}->{"sumV"},
         $::lists{$tb}->{"evals"}->{$evalList}->{"sumF"});
        if ($id eq $::FileIdFP) {
            checkListFP(@args);
        } else {
            checkList(@args);
        }

        # save datafile for this evaluation list
        my $datFile = $prodEval . "." . $::FileExtResult;
        copy($TempDir.$::sep.$datFile, $EvalDir.$::sep.$datFile) or do {
            ::log("ERROR saving $datFile in $EvalDir: $!\n", 3);
            ::log("      Copy data from $TempDir by hand\n\n", 3);
        };

        if ($::debug) {
            ::log("...evaluation for list $evalList done\n", 3);
        }
    } # end of foreach LOOP_EL

    # return successful termination
    return 1;
}




# same as 'doList' but for the list of all archives for packed testbeds
sub doArchiveList {
    # some data that must be passed by the caller
    my $os = $_[0];        # name of current operating system
    my $tb = $_[1];        # name of current testbed
    my $prod = $_[2];      # this is the 3-char code of the current product
    my $checkFile = $_[3]; # name of the file for check data
    my $prodPart = $_[4];  # this is the string that is used as filename
                           # prefix for this evaluation. This is mostly
                           # equal to $prod but can also be different,
                           # eg. $prod="FPR" but $ProdPart="FPRZIP"
    my $EvalDir = $::EvalBase . $::sep . $os . $::sep . $tb;
    my $TempDir = $::TempBase . $::sep . $os . $::sep . $tb;

    my $evalList = $::lists{$tb}->{"packed"}->{"archlist"};
    if ($::debug) {
        ::log("starting evaluation for list $evalList (archives)...\n", 3);
    }

    # there is no ID for archives evaluation
    my $prodEval = $prodPart;

    # process normalized logfile - file level evaluation
    if ($::debug) {
        ::log("doing file level evaluation\n", 3);
    }
    my @files = ($prodEval . "." . $::FileExtEval,
                 $::ListDir . $::sep . $evalList,
                 $prodEval . "." . $::FileExtDetVir,
                 $prodEval . "." . $::FileExtUnrId,
                 $prodEval . "." . $::FileExtMultInf,
                 $prodEval . "." . $::FileExtMultScan,
                 $prodEval . "." . $::FileExtFilteredAno);
    my $anoRef = \%{$::lists{$tb}->{"packed"}->{"ano"}};
    if ($::debug) {
        ::log("args = @files $anoRef\n", 4);
    }
    eval_files(@files, $anoRef) or do {
        # more error details should be given in the subroutine
        ::log("ERROR: evaluating files failed\n", 3);
        return 0;
    };

    # the last step is just counting of the results from the
    # previous step to get the total sums for viruses and samples
    if ($::debug) {
        ::log("counting samples\n", 3);
    }
    my @args = ($prodEval . "." . $::FileExtDetVir,
                $prodEval . "." . $::FileExtResult,
                $::lists{$tb}->{"packed"}->{"sumV"},
                $::lists{$tb}->{"packed"}->{"sumF"},
                $prod,
                "");
    eval_variants(@args) or do {
        # more error details should be given in the subroutine
        return 0;
    };

    if ($::debug) {
        ::log("checking results\n", 3);
    }

    @args = ($tb, $evalList, $prodEval, $checkFile,
             $::lists{$tb}->{"packed"}->{"sumV"},
             $::lists{$tb}->{"packed"}->{"sumF"});
    checkList(@args);

    # save datafile for this evaluation list
    my $datFile = $prodEval . "." . $::FileExtResult;
    copy($TempDir.$::sep.$datFile, $EvalDir.$::sep.$datFile) or do {
        ::log("ERROR saving $datFile in $EvalDir: $!\n", 3);
        ::log("      Copy data from $TempDir by hand\n\n", 3);
    };

    if ($::debug) {
        ::log("    ...evaluation for list $evalList (archives) done\n", 3);
    }
    # return successful termination
    return 1;
}


# return a true value
1;