# Evalcore.pm
# contains core functions for test evaluation
#
# v1.09, 2004-04-05
# Copyright (c) Michel Messerschmidt 2002-2004
#
package Avtc::Evalcore;
use strict;
use Exporter;
use File::Copy;
use File::Basename;
use Cwd;
use IO::File;

@Avtc::Evalcore::ISA = qw(Exporter);
@Avtc::Evalcore::EXPORT = qw(normalizeLogFinal eval_df eval_df_Packed eval_mis
                             eval_mis_Packed eval_files eval_variants
                             eval_sbo eval_postscan eval_postscan_Packed
                             filterLogFP filterPacked filterPackedOk
                             );


sub normalizeLogFinal {
    my $in = $_[0];     # input file in normalized format - may contain
                        # the same file multiple times, is not sorted
                        # and may contain excluded samples
    my $out = $_[1];    # sorted, filtered, normalized logfile - this may
                        # still contain the same file multiple times
    my $outExc = $_[2]; # report all excluded samples here
    my @excList = @{$_[3]}; # all excluded samples, given by reference

    my $line = "";       # holds the current line from input file
    my $fullPath = "";   # file name with full path of current sample
    my $excluded = 0;
    my $excSample;

    # create temporary filenames
    my $tempRand = int(rand(8999)) + 1000;
    my $temp = $::TempBase . $::sep . "nrm_" . $tempRand . ".tmp";

    if ($::debug > 2) {
        ::log("exclude filter: in=$in sorted=$temp out=$out\n", 6);
    }
    # sort the results
    ::mysort($in, $temp, 0) or do {
        return 0;
    };

    if ($::debug) {
        ::log("filtering excluded samples...\n", 4);
    }
    # now filter all excluded samples from the sorted logfile
    if ($::debug > 2) {
        ::log("exclude filter: opening$temp\n", 6);
    }
    open(IN, "< $temp") or do {
        ::log("ERROR opening $temp: $!\n", 4);
        return 0;
    };
    binmode(IN);
    if ($::debug > 2) {
        ::log("exclude filter: opening $out\n", 6);
    }
    open(OUT, "> $out") or do {
        ::log("ERROR opening $out: $!\n", 4);
        close(IN);
        return 0;
    };
    binmode(OUT);
    if ($::debug > 2) {
        ::log("exclude filter: opening $outExc\n", 6);
    }
    open(OUTEXC, "> $outExc") or do {
        close(IN);
        close(OUTDF);
        ::log("ERROR opening $outExc: $!\n", 4);
        return 0;
    };
    binmode(OUTEXC);

    if ($::debug > 2) {
        ::log("exclude filter: main loop\n", 6);
    }
    while (defined($line = <IN>)) {
        $line = ::chommp($line);
        ($fullPath, undef) = split(" ", $line, 2);
        foreach $excSample (@excList) {
            if ($fullPath =~ /^\Q$excSample\E/) {
                $excluded = 1;
                last;
            }
        }
        if (not $excluded) {
            # not filtered - write output line
            print(OUT $line, $::EOL);
        } else {
            # collect all excluded lines in another file
            print(OUTEXC $line, $::EOL);
        }
        $excluded = 0;
    }

    close(IN);
    close(OUT);
    close(OUTEXC);

    # delete temp file
    unlink($temp) or do {
        ::log("ERROR deleting temporary file $temp\n", 4);
    };
    return 1;
}


# this creates the list of all detected files
sub eval_df {
    my $in = $_[0];     # input file in normalized format - can contain
                        # the same file multiple times
    my $outDF = $_[1];  # report all unique filenames here - NOT in normalized
                        # format but only the filenames (incl. path)

    my $return = -1;    # return value: -1 means error,
                        #               values >= 0 mean success
    my $filteredNr = 0; # number of filtered lines
    my $line = "";      # holds the current line from detected files list
    my $fullPath = "";  # file name with full path of current sample
    my $lastPath = "";  # file name with full path of previous sample
    my $excluded = 0;
    my $excSample;

    if ($::debug) {
        ::log("creating 'detected files' list...\n", 4);
    }

    open(IN, "< $in") or do {
        ::log("ERROR opening $in: $!\n", 4);
        return $return;
    };
    binmode(IN);
    open(OUTDF, "> $outDF") or do {
        ::log("ERROR opening $outDF: $!\n", 4);
        close(IN);
        return $return;
    };
    binmode(OUTDF);

    # get the first line from the logfile
    # this is done separately to be able to warn about empty input files
    $line = <IN>;
    $line = ::chommp($line);
    # tests if input file is empty
    if (not $line) {
        # no input => no output (don't write anything);
        ::log("WARNING: found empty input file $in\n", 4);
        close(IN);
        close(OUTDF);
        $return = $filteredNr;
        return $return;
    }
    ($fullPath, undef) = split(" ", $line, 2);
    # write 'detected files' line
    print(OUTDF $fullPath, $::EOL);
    $lastPath = $fullPath;

    # get all other lines
    while (defined($line = <IN>)) {
        $line = ::chommp($line);
        ($fullPath, undef) = split(" ", $line, 2);
        if ($fullPath ne $lastPath) {
            # write 'detected files' line
            print(OUTDF $fullPath, $::EOL);
            $lastPath = $fullPath;
        } else {
            # this is a duplicate line
            $filteredNr++;
        }
    }

    close(IN);
    close(OUTDF);
    # return number of discarded lines
    $return = $filteredNr;
    return $return;
}


# this creates the list of all detected files
sub eval_df_Packed {
    my $prod = $_[0];
    my $outDF = $prod . "." . $::FileExtDetFiles;
    my $archDF = $prod . $::FileExtArchiveFiles . "." . $::FileExtDetFiles;

    # since the normalized logfile for archives already contains ALL archives
    # where at least one sample in it was detected (even if the archive file
    # itself wasn't reported), the detected-files-list for archives is exactly
    # what we want here.
    copy($archDF, $outDF) or do {
        ::log("ERROR: copying $archDF to $outDF failed: $!\n", 4);
    };
}


# this creates a list of all missed files (everything not found as infected)
sub eval_mis {
    # both input files must be sorted the same way !!!
    my $inAll = $_[0];  # dirlist (contains all tested files)
    my $inDF = $_[1];   # list of all detected files
    my $outMis = $_[2]; # list of all files that are not detected, either ok
                        # or not reported at all (created here)
    my $outAF = $_[3];  # list of all anomaly files (excluded from outMis)
    my $outGF = $_[4];  # if there are files in inDF, that are not found in
                        # in the dirlist they are collected here
    my %anoList = %{$_[5]}; # all anomalies (files & nr), given by reference
    # set correct syntax for logfile parsing
    if ($::configSyntax eq $::OS_WINDOWS) {
        fileparse_set_fstype("MSWin32");
    } else {
        fileparse_set_fstype("Unix");
    }

    if ($::debug > 0) {
        ::log("creating 'missed samples' list...\n", 4);
    }
    if ($::debug > 2) {
        ::log("args: dirlist=$inAll\n", 5);
        ::log("        inD_f=$inDF\n", 5);
        ::log("    OutMissed=$outMis\n", 5);
        ::log("       OutA_f=$outAF\n", 5);
        ::log("       OutG_f=$outGF\n", 5);
    }

    open(INALL, "< $inAll") or do {
        ::log("ERROR opening $inAll: $!\n", 4);
        return 0;
    };
    binmode(INALL);
    open(INDET, "< $inDF") or do {
        ::log("ERROR opening $inDF: $!\n" ,4);
        close(INALL);
        return 0;
    };
    binmode(INDET);
    open(OUTMIS, "> $outMis") or do {
        ::log("ERROR opening $outMis: $!\n", 4);
        close(INALL);
        close(INDET);
        return 0;
    };
    binmode(OUTMIS);
    open(OUTAF, "> $outAF") or do {
        ::log("ERROR opening $outAF: $!\n", 4);
    close(INALL);
    close(INDET);
    close(OUTMIS);
        return 0;
    };
    binmode(OUTAF);
    open(OUTGF, "> $outGF") or do {
        ::log("ERROR opening $outGF: $!\n", 4);
    close(INALL);
    close(INDET);
    close(OUTMIS);
    close(OUTAF);
        return 0;
    };
    binmode(OUTGF);

    my @misFiles = ();
    my @anoFiles = ();
    my $listLine;
    my $samples = 0;
    my $variant;
    my $lastVariant = "";
    my $detSamples = 0;
    my $detVariant;
    # get first line from detected files list
    my $detLine = <INDET>;
    if (defined($detLine)) {
        $detLine = ::chommp($detLine);
        # count samples for each variant - this is required to exclude
        # anomaly files
        (undef, $detVariant) = fileparse($detLine);
        chop($detVariant);
        $detSamples = 1;
    }
    OUTERLIST: while (defined($listLine = <INALL>)) {
        $listLine = ::chommp($listLine);
        # count samples for each variant - this is required to exclude
        # anomaly files
        (undef, $variant) = fileparse($listLine);
        chop($variant);
        if ($variant eq $lastVariant) {
            $samples++;
        } else {
            # Add all anomaly files to missed sample list
            # iff none of them was detected.
            # If at least one anomaly sample was detected,
            # all of them are treated as detected.
            if ($detSamples < 1) {
                push @misFiles, @anoFiles;
            } else {
                # These files are not counted as missed samples
                # but added to the list of filtered anomalies
                foreach my $file (@anoFiles) {
                    print(OUTAF $file, $::EOL);
                }
            }
            foreach my $file (@misFiles) {
                print(OUTMIS $file, $::EOL);
            }
            @misFiles = ();
            @anoFiles = ();

            if (defined($detVariant) and $detVariant eq $variant) {
                $detSamples = 1;
            } else {
                $detSamples = 0;
            }
            $samples = 1;
            $lastVariant = $variant;
        }
        INNERLIST: while (defined($detLine)) {
            if ($detLine eq $listLine) {
                # look at next dirList-line if this one is found in det-list
                $detLine = <INDET>;
                if (defined($detLine)) {
                    $detLine = ::chommp($detLine);
                    # count samples for each variant - this is required to
                    # exclude anomaly files
                    (undef, $detVariant) = fileparse($detLine);
                    chop($detVariant);
                    if ($detVariant eq $variant) {
                        $detSamples++;
                    }
                } else {
                    $detSamples = 0;
                }
                next OUTERLIST;
            } elsif ($detLine gt $listLine) {
                # Found some missed samples...
                # Exclude anomaly files from missed files list, iff
                # at least one of them was detected
                foreach my $ano (keys %anoList) {
                    if ($listLine =~ /^\Q$ano\E/) {
                        if ($::debug) {
                            ::log("found anomaly file: $listLine\n", 4);
                        }
                        push @anoFiles, $listLine;
                        next OUTERLIST;
                    }
                }
                push @misFiles, $listLine;
                next OUTERLIST;
            }
            # Since both lists are sorted, and dirlist is the complete list
            # a missing line in det-list results in the next det-lines being
            # greater (in terms of ASCII order) than the dirlist-line.
            # If the det-line is less that the dirlist line, there must be
            # some wrong lines in the det-list (scanner keys can easily
            # produce garbage entries that find their way to the det-list).
            # Collect all garbage lines in gf-file:
            print(OUTGF $detLine, $::EOL);
            $detLine = <INDET>;
            if (defined($detLine)) {
                $detLine = ::chommp($detLine);
                # count samples for each variant - this is required to exclude
                # anomaly files
                (undef, $detVariant) = fileparse($detLine);
                chop($detVariant);
                if ($detVariant eq $variant) {
                    $detSamples++;
                }
            } else {
                $detSamples = 0;
            }
        }
        # If detLine is not defined we're at the end of the det-list
        # (and the remaining files are all missed files)
        print(OUTMIS $listLine, $::EOL);
    }
    foreach my $file (@misFiles) {
        print(OUTMIS $file, $::EOL);
    }

    close(INALL);
    close(INDET);
    close(OUTMIS);
    close(OUTAF);
    close(OUTGF);
    return 1;
}


# this creates a list of all missed files (everything not found as infected)
sub eval_mis_Packed {
    my $prod = $_[0];
    my $tb = $_[1];
    my %packPrg = %{$_[2]}; # contains list of compression programs and
                            # their archive filenames
    my $outMis = $prod . "." . $::FileExtMissedFiles;
    my $archMis = $prod . $::FileExtArchiveFiles . "." . $::FileExtMissedFiles;
    # create temporary filenames
    my $tempRand = int(rand(8999)) + 1000;
    my $temp = $::TempBase . $::sep . "misp" . $tempRand . ".tmp";
    # set correct syntax for logfile parsing
    if ($::configSyntax eq $::OS_WINDOWS) {
        fileparse_set_fstype("MSWin32");
    } else {
        fileparse_set_fstype("Unix");
    }

    if ($::debug) {
        ::log("evaluating missed samples for packed testbeds...\n", 4);
    }

    # the missed files list for packed testbeds is the summary of the missed
    # archive and all archives where at least one file from the contents is
    # missed.
    # First get the archives part
    copy($archMis, $temp) or do {
        ::log("ERROR: copying $archMis to $temp failed: $!\n", 4);
    };

    # now append all missed files list from all packers and sort it
    open(TEMP, ">> $temp") or do {
        ::log("ERROR appending to $temp: $!\n", 4);
        return 0;
    };
    binmode(TEMP);
    my $pack;
    foreach $pack (sort keys %packPrg) {
        my $mis = $prod . $pack . "." . $::FileExtMissedFiles;
        if (-s $mis) {
            open(INMIS, "< $mis") or do {
                ::log("ERROR opening $mis: $!\n", 4);
                close(TEMP);
                return 0;
            };
            binmode(INMIS);

            my $line;
            my $lastPath = "";
            while (defined($line = <INMIS>)) {
                $line = ::chommp($line);
                my $tempPath;
                # cut filename from path and ...
                (undef, $tempPath) = fileparse($line);
                # ... append archive name to path instead
                $tempPath .= $packPrg{$pack}{"filename"};
                # don't write more than one line per archive
                if ($tempPath ne $lastPath) {
                    $lastPath = $tempPath;
                    print(TEMP $tempPath, $::EOL);
                }
            }
            close(INMIS);
        }
    }
    close(TEMP);

    # sort missed files list and eliminate identical lines
    ::mysort($temp, $outMis, 1) or do {
        return 0;
    };

    # delete temp files
    unlink($temp) or do {
        ::log("ERROR deleting temporary file $temp\n", 4);
    };
    return 1;
}



# this produces a sum of detected files for each directory (= virus-variant)
# lists of multiple infections/scans are produced too (in separate files)
sub eval_files {
    # both input files must be sorted the same way !!!
    my $in = $_[0];     # list of all files to be evaluated
    my $inVar = $_[1];  # evaluation list (variantlist) to compare to
    my $out = $_[2];    # evaluation list enhanced with result data
                        # (appended in each line)
    my $outUI = $_[3];  # report all unreliable identifications here
    my $outMI = $_[4];  # report all multiple infected samples here
    my $outMS = $_[5];  # report all multiple scanned samples here
    my $outAF = $_[6];  # report all anomaly files filtered out
    my %anoList = %{$_[7]}; # all anomalies (files & nr), given by reference

    my $line = "";       # holds the current line from detected files list
    my $varLine = "";    # holds the current line from variantlist
    my $samples = 0;     # counts nr. of samples/files from a virus-variant/dir
    my $unreliable = 0;  # is set if a virus-variant isn't identified correct
    my $virName = "";    # reported virus name
    my $fullPath = "";   # file name with full path of current sample
    my $path = "";       # only the path (= name of this virus variant)
    my $fileName = "";   # only file name of current sample
    my $lastVirName = "";  # and the same for the previous line...
    my $lastPath = "";
    my $lastFileName = "";
    my $varPath = "";    # Name of virus-variant in variantlist
    my $varNr = 0;       # Number of samples of this virus-variant
    # set correct syntax for logfile parsing
    if ($::configSyntax eq $::OS_WINDOWS) {
        fileparse_set_fstype("MSWin32");
    } else {
        fileparse_set_fstype("Unix");
    }

    if ($::debug) {
        ::log("evaluating files...\n", 4);
    }

    # open files to read and write line by line to avoid big arrays in memory
    open(IN, "< $in") or do {
        ::log("ERROR opening $in: $!\n", 4);
        close(IN);
        return 0;
    };
    binmode(IN);
    open(INVAR, "< $inVar") or do {
        ::log("ERROR opening $inVar: $!\n", 4);
        close(IN);
        close(INVAR);
        return 0;
    };
    binmode(INVAR);
    open(OUT, "> $out") or do {
        ::log("ERROR opening $out: $!\n", 4);
        close(IN);
        close(INVAR);
        close(OUT);
        return 0;
    };
    binmode(OUT);
    open(OUTUI, "> $outUI") or do {
        ::log("ERROR opening $outUI: $!\n", 4);
        close(IN);
        close(INVAR);
        close(OUT);
        close(OUTUI);
        return 0;
    };
    binmode(OUTUI);
    open(OUTMI, "> $outMI") or do {
        ::log("ERROR opening $outMI: $!\n", 4);
        close(IN);
        close(INVAR);
        close(OUT);
        close(OUTUI);
        close(OUTMI);
        return 0;
    };
    binmode(OUTMI);
    open(OUTMS, "> $outMS") or do {
        ::log("ERROR opening $outMS: $!\n", 4);
        close(IN);
        close(INVAR);
        close(OUT);
        close(OUTUI);
        close(OUTMI);
        close(OUTMS);
        return 0;
    };
    binmode(OUTMS);
    open(OUTAF, "> $outAF") or do {
        ::log("ERROR opening $outAF: $!\n", 4);
        close(IN);
        close(INVAR);
        close(OUT);
        close(OUTUI);
        close(OUTMI);
        close(OUTMS);
        close(OUTAF);
        return 0;
    };
    binmode(OUTAF);

    # get the first line from input file
    EF_PRE_LOOP: while (defined($line = <IN>)) {
        my $ano;
        $line = ::chommp($line);
        # tests if input file is empty ("0" will never be a valid dirname
        #                               - there's always a drive letter)
        if (not $line) {
            # no input -> no output (don't write anything);
            ::log("WARNING: found empty input file $in\n", 4);
            close(IN);
            close(OUT);
            close(OUTUI);
            close(OUTMI);
            close(OUTMS);
            return 1;
        }
        # get path, filename and reported virusname from line
        ($fullPath, $virName) = split(" ", $line);
        ($fileName, $path) = fileparse($fullPath);
        # remove trailing backslash from path (would cause errors otherwise)
        chop($path);

        # filter anomaly files before counting this sample
        foreach $ano (keys %anoList) {
            # since this is the first sample of a virus, it is only filtered if
            # the maximum number of counted samples for this virus is 0
            if (($line =~ /^\Q$ano\E/)
                    and (0 >= $anoList{$ano})) {
                if ($::debug) {
                    ::log("filtered anomaly file: $line\n", 4);
                }
                # this file is not counted as sample
                # but added to the list of filtered anomalies
                print(OUTAF $fullPath, $::EOL);
                next EF_PRE_LOOP;
            }
        }
        $samples = 1;
        if ($virName eq $::VirusNameUI) {
            # it's an unreliable identification (no exact identification)
            $unreliable = 1;
            printf(OUTUI "%-40s %-15s%s",$path,$::VirusNameUI, $::EOL);
        }
        $lastVirName = $virName;
        $lastPath = $path;
        $lastFileName = $fileName;
        # exit this loop immediately, if the first line is no anomaly file
        last EF_PRE_LOOP;
     }

    #get the first line from variant list
    $varLine = <INVAR>;
    ($varPath, $varNr) = split(" ", ::chommp($varLine));
    # print all leading virus-variants in variantlist that were not found
    while (($varPath ne $path) and defined($varLine = <INVAR>)) {
        if ($::debug > 2) {
            ::log("variant not found: Var-path=$varPath  - Nr=$varNr\n", 4);
        }
        printf(OUT "%s %s 0 0%s", $varPath, $varNr, $::EOL);
        # go on with next line from variantlist
        ($varPath, $varNr) = split(" ", ::chommp($varLine));
    }
    ### ATTENTION: don't use "if (eof(INVAR))" here (or in similar loops),
    ### because this checks the current state of the opened filehandle, not
    ### the state from the last read call ("<INVAR>"). And if the last
    ### read call already read the last line, this last line is not processed
    if (not defined($varLine)) {
        # seems like we found nothing useful
        # there can be no sample of a valid virus that's not in the variantlist
        close(IN);
        close(INVAR);
        close(OUT);
        close(OUTUI);
        close(OUTMI);
        close(OUTMS);
        close(OUTAF);
        return 1;
    }

    # get all other lines
    EF_LOOP: while (defined($line = <IN>)) {
        my $ano;
        $line = ::chommp($line);
        ($fullPath, $virName) = split(" ", $line);
        ($fileName, $path) = fileparse($fullPath);
        chop($path);


        if ($path eq $lastPath) {
            # it's the same virus-variant
            if ($fileName eq $lastFileName) {
                # it's the same sample
                if ($virName eq $lastVirName) {
                    # must be scanned multiple times
                    print(OUTMS $fullPath, $::EOL);
                } else {
                    # multiple infections were reported for the same file
                    printf(OUTMI "%-40s %-15s %-15s%s",$fullPath,
                                 $lastVirName,$virName, $::EOL);
                    $lastVirName = $virName;
                }
            } else {
                # it's another sample of the same virus-variant
                if ($virName ne $lastVirName) {
                    # it's an unreliable identification (different virusnames
                    # were reported for the same virus-variant)
                    $unreliable = 1;
                    printf(OUTUI "%-40s %-15s %-15s%s",$path,
                                 $lastVirName, $virName, $::EOL);
                }
                if ($virName eq $::VirusNameUI) {
                    # it's an unreliable identification (not exact identified)
                    $unreliable = 1;
                    printf(OUTUI "%-40s %-15s%s",$path,$::VirusNameUI, $::EOL);
                }
                $lastFileName = $fileName;
                $lastVirName = $virName;
                # filter anomaly files before counting this sample
                foreach $ano (keys %anoList) {
                    if (($line =~ /^\Q$ano\E/)
                        and ($samples >= $anoList{$ano})) {
                        if ($::debug) {
                            ::log("filtered anomaly file: $line\n", 4);
                        }
                        # this file is not counted as sample
                        # but added to the list of filtered anomalies
                        print(OUTAF $fullPath, $::EOL);
                        next EF_LOOP;
                    }
                }
                $samples++;
            }
        } else {
            # it's the next virus-variant
            # filter anomaly files before counting this sample
            foreach $ano (keys %anoList) {
                # since this is the first sample of a virus, it's only filtered
                # if the maximum number of counted samples for this virus is 0
                if (($line =~ /^\Q$ano\E/)
                    and (0 >= $anoList{$ano})) {
                    if ($::debug) {
                        ::log("filtered anomaly file: $line\n", 4);
                    }
                    # this file is not counted as sample
                    # but added to the list of filtered anomalies
                    print(OUTAF $fullPath, $::EOL);
                    next EF_LOOP;
                }
            }
            # join with data from variantlist
            printf(OUT "%s %s %s %s%s", $varPath, $varNr, $samples,
                   $unreliable, $::EOL);
            if (defined($varLine = <INVAR>)) {
                ($varPath, $varNr) = split(" ", ::chommp($varLine));
                # print all virus-variants not found
                while (($varPath ne $path) and defined($varLine = <INVAR>)) {
                    if ($::debug > 2) {
                        ::log("variant not found: Var-path=$varPath  - Nr=".
                              $varNr."\n", 4);
                    }
                    printf(OUT "%s %s 0 0%s", $varPath, $varNr, $::EOL);
                    ($varPath, $varNr) = split(" ", ::chommp($varLine));
                }
            }
            if (not defined($varLine)) {
                # there can be no sample of a valid virus that's not in the
                # variantlist, therefore the list should be cut here
                # (or we could get really strange results)
                close(IN);
                close(INVAR);
                close(OUT);
                close(OUTUI);
                close(OUTMI);
                close(OUTMS);
                close(OUTAF);
                return 1;
            }
            # go on with the next virus-variant
            $lastPath = $path;
            $lastFileName = $fileName;
            $lastVirName = $virName;
            $samples = 1;
            $unreliable = 0;
            if ($virName eq $::VirusNameUI) {
                # it's an unreliable identification (no exact identification)
                $unreliable = 1;
                printf(OUTUI "%-40s %-15s%s",$path,$::VirusNameUI, $::EOL);
            }
        }
    }
    # don't forget the last line / virus-variant
    printf(OUT "%s %s %s %s%s", $varPath, $varNr, $samples, $unreliable,
       $::EOL);
    close(IN);
    close(INVAR);
    close(OUT);
    close(OUTUI);
    close(OUTMI);
    close(OUTMS);
    close(OUTAF);
    return 1;
}



# this produces the final result of detected files and viruses
sub eval_variants {
    my $in = $_[0];         # input file
    my $out = $_[1];        # output file
    my $nrViruses = $_[2];  # total number of viruses
    my $nrSamples = $_[3];  # total number of samples
    my $prod = $_[4];       # product code
    my $id = $_[5];         # type of evaluation list
    # init variables
    my $line = "";       # holds the current line from input file
    my $virusName = 0;    # directory / virus-name
    my $samples = 0;      # number of samples in this directory
    my $samplesFound = 0; # number of samples in this dir found by the product
    my $unIdent = 0;      # was this virus unreliable identified (1=yes,0=no)
    my $nrSamplesFound = 0;  # total number of samples found by the product
    my $nrVirusesFound = 0; # total number of viruses found by the product
    my $nrUnIdent =0;     # total number of unreliable identified viruses
    my $nrUnDet = 0;      # total number of unreliable detected viruses

    if ($::debug) {
        ::log("evaluating viruses/variants...\n", 4);
    }

    # open files to read and write line by line to avoid big arrays in memory
    # (may be too much for our older machines)
    open(IN, "< $in") or do {
        ::log("ERROR opening $in: $!\n", 4);
        close(IN);
        return 0;
    };
    binmode(IN);
    open(OUT, "> $out") or do {
        ::log("ERROR opening $out: $!\n", 4);
        close(IN);
        close(OUT);
        return 0;
    };
    binmode(OUT);

    while (defined($line = <IN>)) {
        $line = ::chommp($line);
        ($virusName, $samples, $samplesFound, $unIdent) = split(" ", $line);
        # ignore lines with no samples (should not occur)
        if ($samples == 0) {
            if ($::debug) {
                ::log("WARNING: evallist contains 0 samples for '".$virusName.
                      "'\n", 4);
            }
            next;
        }
        $nrSamplesFound += $samplesFound;
        if ($samplesFound == $samples) {
            $nrVirusesFound++;
            if ($unIdent) {
                $nrUnIdent++;
            }
        } elsif ($samplesFound > 0) {
            $nrVirusesFound++;
            $nrUnDet++;
        }
    }

    # calculate percent values
    my $percVirusesFound = $nrVirusesFound * 100 / $nrViruses;
    my $percUnIdent = $nrUnIdent * 100 / $nrViruses;
    my $percUnDet = $nrUnDet * 100 / $nrViruses;
    my $percSamplesFound = $nrSamplesFound * 100 / $nrSamples;

    # mark rounded percent values (insert ~ as decimal, if value is rounded)
    $percVirusesFound = sprintf("%5.1f", $percVirusesFound);
    if ($percVirusesFound eq "100.0" and $nrVirusesFound < $nrViruses) {
        $percVirusesFound = "100.~";
    } elsif ($percVirusesFound eq "  0.0" and $nrVirusesFound > 0) {
        $percVirusesFound = "  0.~";
    }

    $percUnIdent = sprintf("%5.1f", $percUnIdent);
    if ($percUnIdent eq "100.0" and $nrUnIdent < $nrViruses) {
        $percUnIdent = "100.~";
    } elsif ($percUnIdent eq "  0.0" and $nrUnIdent > 0) {
        $percUnIdent = "  0.~";
    }

    $percUnDet = sprintf("%5.1f", $percUnDet);
    if ($percUnDet eq "100.0" and $nrUnDet < $nrViruses) {
        $percUnDet = "100.~";
    } elsif ($percUnDet eq "  0.0" and $nrUnDet > 0) {
        $percUnDet = "  0.~";
    }

    $percSamplesFound = sprintf("%5.1f", $percSamplesFound);
    if ($percSamplesFound eq "100.0" and $nrSamplesFound < $nrSamples) {
        $percSamplesFound = "100.~";
    } elsif ($percSamplesFound eq "  0.0" and $nrSamplesFound > 0) {
        $percSamplesFound = "  0.~";
    }

    if ($::debug) {
        ::log("writing result file $out\n", 4);
    }
    if ($id eq "F") {
        # false positives result have an other format
        printf(OUT "%s      %6d %s%s", $prod, $nrSamplesFound,
                   $percSamplesFound, $::EOL);
    } else {
        printf(OUT "%s      %6d %s %5d %s %5d %s %6d %s%s", $prod,
                   $nrVirusesFound, $percVirusesFound, $nrUnIdent,
                   $percUnIdent, $nrUnDet, $percUnDet,
                   $nrSamplesFound, $percSamplesFound, $::EOL);
    }
    close(IN);
    close(OUT);
    return 1;
}



# get a real ok_files list (sorted and without identical lines)
sub eval_sbo {
    my $in = $_[0];     # input file containing filenames - can contain
                        # the same file multiple times
    my $out = $_[1];    # report all unique filenames here - NOT in normalized
                        # format but only the filenames (incl. path)
    my @excList = @{$_[2]}; # all excluded samples, given by reference

    my $return = -1;    # return value: -1 means error,
                        #               values >= 0 mean success
    my $filteredNr = 0; # number of filtered lines
    my $inLine = "";    # holds the current line from detected files list
    my $lastLine = "";  # file name with full path of previous sample
    my $excluded = 0;
    my $excSample;
    # create temporary filenames
    my $tempRand = int(rand(8999)) + 1000;
    my $temp = $::TempBase . $::sep . "of_s" . $tempRand . ".tmp";
    # delete existing (old) temp files
    if (-e $temp) {
        unlink($temp) or do {
            ::log("ERROR deleting teporary file $temp\n", 4);
        };
    }
    # sort 'scanned but okay' files list
    ::mysort($in, $temp, 0) or do {
        ::log("ERROR sorting $in\n", 4);
        return $return;
    };

    if ($::debug) {
        ::log("creating 'okay files' list...\n", 4);
    }
    open(IN, "< $temp") or do {
        ::log("ERROR opening $temp: $!\n", 4);
        return $return;
    };
    binmode(IN);
    open(OUT, "> $out") or do {
        ::log("ERROR opening $out: $!\n", 4);
        close(IN);
        return $return;
    };
    binmode(OUT);

    while (defined($inLine = <IN>)) {
        $inLine = ::chommp($inLine);
        if ($inLine ne $lastLine) {
            EXC_LOOP: foreach $excSample (@excList) {
                if ($inLine =~ /^\Q$excSample\E/) {
                    $excluded = 1;
                    last EXC_LOOP;
                }
            }
            if (not $excluded) {
                # write 'okay files' line
                print(OUT $inLine, $::EOL);
                $lastLine = $inLine;
            } else {
                # this line is filtered out
                $filteredNr++;
            }
            $excluded = 0;
        } else {
            # silenty discard duplicate lines
            # but count the number of discarded lines
            $filteredNr++;
        }
    }
    close(IN);
    close(OUT);

    # delete temp files
    unlink($temp) or do {
        ::log("ERROR deleting temporary file $temp\n", 4);
    };

    # return number of discarded lines
    $return = $filteredNr;
    return $return;
}


# create postscan list and list of all scanned files
sub eval_postscan {
    my $prod = $_[0];     # filename prefix
    my $dirList = $_[1];  # list of all files in this testbed (directory)
    my $isFP = $_[2];     # is this a false positive testbed ?

    # list of all detected samples
    my $df = $prod . "." . $::FileExtDetFiles;
    # list of all scanned but undetected samples
    my $of = $prod . "." . $::FileExtOKFiles;
    # list of all scanned files (created here)
    my $sf = $prod . "." . $::FileExtScanFiles;
    # list of all files reported as infected and as okay (created here)
    my $ino = $prod . "." . $::FileExtInfAndOK;
    # postscan list (created here)
    my $ps = $prod . "." . $::FileExtPostscan;
    # list of all garbage files (created here)
    my $gf = $prod . "." . $::FileExtGarbage;

    if ($::debug) {
        ::log("creating postscan list...\n", 4);
    }
    # create temporary filenames
    my $tempRand = int(rand(8999)) + 1000;
    my $temp = $::TempBase . $::sep . "sf_u" . $tempRand . ".tmp";
    my $temp2 = $::TempBase . $::sep . "sf_s" . $tempRand . ".tmp";
    # delete existing (old) temp files
    if (-e $temp) {
        unlink($temp) or do {
            ::log("ERROR deleting teporary file $temp\n", 4);
        };
    }
    if (-e $temp2) {
        unlink($temp2) or do {
            ::log("ERROR deleting teporary file $temp2\n", 4);
        };
    }

    #create list of scanned files (detected and undetected samples)
    copy($df, $temp) or do {
        ::log("ERROR copying $df to $temp\n", 4);
        ::log("The postscan list wasn't created correctly.\n", 4);
        return 0;
    };

    if ($isFP) {
        # additionally append all detected false positive files
        my $dfp = $prod . "." . $::FileExtDetFPFiles;
        ::appendFile($temp, $dfp) or do {
            ::log("ERROR appending $dfp to $temp\n", 4);
            ::log("The postscan list wasn't created correctly.\n", 4);
            return 0;
        };
    }

    ::appendFile($temp, $of) or do {
        ::log("ERROR appending $of to $temp\n", 4);
        ::log("The postscan list wasn't created correctly.\n", 4);
        return 0;
    };
    # sort scanned files list
    ::mysort($temp, $temp2, 0) or do {
        return 0;
    };

    # find all identical lines - these are files reported as infected AND as
    # okay and were therefore in both .d_f and .o_f
    # this eliminates all identical lines in .s_f and stores them in .ino
    open(INTEMP, "< $temp2") or do {
        close(INTEMP);
        ::log("ERROR opening $temp2: $!\n", 4);
        return 0;
    };
    binmode(INTEMP);
    open(OUTSF, "> $sf") or do {
        close(INTEMP);
        close(OUTSF);
        ::log("ERROR opening $sf: $!\n", 4);
        return 0;
    };
    binmode(OUTSF);
    open(OUTINO, "> $ino") or do {
        close(INTEMP);
        close(OUTSF);
        close(OUTINO);
        ::log("ERROR opening $sf: $!\n", 4);
        return 0;
    };
    binmode(OUTINO);

    my $lastTempLine = "";
    my $tempLine;
    while (defined($tempLine = <INTEMP>)) {
        $tempLine = ::chommp($tempLine);
        if ($tempLine ne $lastTempLine) {
            print(OUTSF $tempLine, $::EOL);
            $lastTempLine = $tempLine;
        } else {
            print(OUTINO $tempLine, $::EOL);
        }
    }
    close(INTEMP);
    close(OUTSF);
    close(OUTINO);

    # join scanned files list with dirlist,
    # (only unmatched files in dirlist are reported)
    open(DIRLIST, "< $dirList") or do {
        ::log("ERROR opening $dirList: $!\n", 4);
        return 0;
    };
    binmode(DIRLIST);
    open(INSF, "< $sf") or do {
        close(DIRLIST);
        ::log("ERROR opening $sf: $!\n", 4);
        return 0;
    };
    binmode(INSF);
    open(OUTGF, "> $gf") or do {
        close(DIRLIST);
        close(INSF);
        ::log("ERROR opening $gf: $!\n", 4);
        return 0;
    };
    binmode(OUTGF);
    open(OUTPS, "> $ps") or do {
        close(DIRLIST);
        close(INSF);
        close(OUTGF);
        close(OUTEXC);
        ::log("ERROR opening $ps: $!\n", 4);
        return 0;
    };
    binmode(OUTPS);
    my $listLine;
    # get first line from s_f
    my $sfLine = <INSF>;
    OUTERLIST: while (defined($listLine = <DIRLIST>)) {
        $listLine = ::chommp($listLine);
        INNERLIST: while (defined($sfLine)) {
            $sfLine = ::chommp($sfLine);
            if ($sfLine eq $listLine) {
                # look at next dirList-line if this one is found in sf-list
                $sfLine = <INSF>;
                next OUTERLIST;
            } elsif ($sfLine ge $listLine) {
                print(OUTPS $listLine, $::EOL);
                next OUTERLIST;
            }
            # since both lists are sorted, and dirlist is the complete list
            # a mising line in sf-list results in the next sf-lines being
            # greater (in terms of ASCII order) than the dirlist-line.
            # If the sf-line is less that the dirlist line, there must be
            # some wrong lines in the sf-list (scanner keys can easily write
            # garbage to the sf-list)
            print(OUTGF $sfLine, $::EOL);
            $sfLine = <INSF>;
        }
        # if sfLine is not defined we're at the end of the s_f-file
        print(OUTPS $listLine, $::EOL);
    }
    close(OUTPS);
    close(OUTGF);
    close(INSF);
    close(DIRLIST);

    # delete temp files
    unlink($temp) or do {
        ::log("ERROR deleting temporary file $temp\n", 4);
    };
    unlink($temp2) or do {
        ::log("ERROR deleting temporary file $temp2\n", 4);
    };

    return 1;
}


# this creates the postscan list for the whole testbed (all packers)
sub eval_postscan_Packed {
    my $prod = $_[0];
    my $tb = $_[1];
    my %packPrg = %{$_[2]}; # contains list of compression programs and
                            # their archive filenames
    my $outFile = $prod . "." . $::FileExtPostscan;
    my $archFile = $prod . $::FileExtArchiveFiles . "." . $::FileExtPostscan;
    # create temporary filename
    my $tempRand = int(rand(8999)) + 1000;
    my $tempFile = $::TempBase . $::sep . "pslu" . $tempRand . ".tmp";
    # set correct syntax for logfile parsing
    if ($::configSyntax eq $::OS_WINDOWS) {
        fileparse_set_fstype("MSWin32");
    } else {
        fileparse_set_fstype("Unix");
    }

    if ($::debug) {
        ::log("creating postscan list for packed testbeds...\n", 4);
    }
    # get all postscans from archive evaluation first
    copy($archFile, $tempFile) or do {
        ::log("ERROR: copying $archFile to $tempFile failed: $!\n", 4);
        ::log("ERROR creating $outFile: $!\n", 4);
        return 0;
    };

    # and now append postscan list from all packer evaluations
    # this requires cutting of the filename
    open(OUT, ">> $tempFile") or do {
        ::log("ERROR opening $tempFile: $!\n", 4);
        return 0;
    };
    binmode(OUT);

    my $pack;
    foreach $pack (sort keys %packPrg) {
        # Only append the postscanlist for a specific packer to the complete
        # postscanlist, if the 'scanned files' list for this packer has
        # non-zero size.
        # This is to prevent the case, that all archive are scanned as OK
        # (therefore no postscans are needed), but since the archive contents
        # may not be not reported, the sf-file for each packer is empty
        # (the sf-file exists in such a case but has zero size).
        my $sfFile = $prod . $pack . "." . $::FileExtScanFiles;
        if (-s $sfFile) {
            my $packFile = $prod . $pack . "." . $::FileExtPostscan;
            open(IN, "< $packFile") or do {
                ::log("ERROR opening $packFile: $!\n", 4);
                ::log("ERROR creating $outFile: $!\n", 4);
                close(OUT);
                return 0;
            };
            binmode(IN);

            my $line;
            my $lastPath = "";
            while (defined($line = <IN>)) {
                $line = ::chommp($line);
                my $tempPath;
                # cut filename from path and ...
                (undef, $tempPath) = fileparse($line);
                # ... append archive name to path instead
                $tempPath .= $packPrg{$pack}{"filename"};
                # don't write more than one line pro archive
                if ($tempPath ne $lastPath) {
                    $lastPath = $tempPath;
                    print(OUT $tempPath, $::EOL);
                }
            }
            close(IN);
        }
    }
    close(OUT);

    # now it is possible that there are multiple identical lines
    # sort postscan list and eliminate identical lines
    ::mysort($tempFile, $outFile, 1) or do {
        return 0;
    };

    # delete temp files
    unlink($tempFile) or do {
        ::log("ERROR deleting temporary file $tempFile\n", 4);
    };
    return 1;
}




# extract data to evaluate with one evallist from logfile containing all data
sub filterLogFP {
    my $prodCode = $_[0]; # filename prefix (= product code)  [FPR / FPRARJ]
    my $inEval = $_[1];   # sample list to decide which lines to filter
                          # [MZ______]
#    my @ids = @{$_[2]};      # list of evaluationlist IDs to filter logs
                             # for them [() / I]
    # must contain a value that is always greater than any line in inLog
    # (when comparing lines)
    my $BIGVALUE = chr(255);
    # create all filenames
    my $inLog = $prodCode . "." . $::FileExtNorm;  # normalized logfile
    my $inFP = $inEval;
    substr($inFP, 2, 1) = $::FileIdFP;  # ID for FP evaluation should be "F"
    # input files
    my $inEvalFile = $::ListDir . $::sep . $inEval . "." . $::FileExtSamples;
    my $inFPFile = $::ListDir . $::sep . $inFP . "." . $::FileExtSamples;
    # output files
    my $outEval = $prodCode . "." . $::FileExtEval;
    my $outFP = $prodCode . "_" . $::FileIdFP . "." . $::FileExtEval;
    # file for remaining lines (for error detection)
    my $outRem = $prodCode . "." . $::FileExtRem;
    if ($::debug) {
        ::log("filtering false positive samples\n", 4);
        ::log("inLogFile = $inLog\n", 5);
        ::log("inEvalFile = $inEvalFile\n", 5);
        ::log("inFPFile = $inFPFile\n", 5);
        ::log("outEvalFile = $outEval\n", 5);
        ::log("outEvalFP = $outFP\n", 5);
        ::log("outEvalRem = $outRem\n", 5);
    }

    open(INLOG, "< $inLog") or do {
        ::log("ERROR opening $inLog: $!\n", 4);
        return 0;
    };
    binmode(INLOG);
    open(INEVAL, "< $inEvalFile") or do {
        close(INLOG);
        ::log("ERROR opening $inEvalFile: $!\n", 4);
        return 0;
    };
    binmode(INEVAL);
    open(INFP, "< $inFPFile") or do {
        close(INLOG);
        close(INEVAL);
        ::log("ERROR opening $inFPFile: $!\n", 4);
        return 0;
    };
    binmode(INFP);
    open(OUTEVAL, "> $outEval") or do {
        close(INLOG);
        close(INEVAL);
        close(INFP);
        ::log("ERROR opening $outEval: $!\n", 4);
        return 0;
    };
    binmode(OUTEVAL);
    open(OUTFP, "> $outFP") or do {
        close(INLOG);
        close(INEVAL);
        close(INFP);
        close(OUTEVAL);
        ::log("ERROR opening $outFP: $!\n", 4);
        return 0;
    };
    binmode(OUTFP);
    open(OUTREM, "> $outRem") or do {
        close(INLOG);
        close(INEVAL);
        close(INFP);
        close(OUTEVAL);
        close(OUTFP);
        ::log("ERROR opening $outRem: $!\n", 4);
        return 0;
    };
    binmode(OUTREM);

    # set correct syntax for logfile parsing
    if ($::configSyntax eq $::OS_WINDOWS) {
        fileparse_set_fstype("MSWin32");
    } else {
        fileparse_set_fstype("Unix");
    }

    my ($logFullPath, $logPath, $logFile);
    my ($evalFullPath, $evalPath, $evalFile);
    my ($fpFullPath, $fpPath, $fpFile);
    # get first line of each input file
    my $logLine = <INLOG>;
    my $evalLine = <INEVAL>;
    my $fpLine = <INFP>;
    my $loopEval = 1;
    if (not defined($evalLine)) {
        $loopEval = 0;
        # if this list is empty, the value must be greater than all other
        # possible values to for correct loop operation below
        $evalLine = $BIGVALUE;
    }
    my $loopFP = 1;
    if (not defined($fpLine)) {
        $loopFP = 0;
        $fpLine = $BIGVALUE;
    }
    FFP_OUTER: while (defined($logLine)) {
        # get all relevant data parts
        $logLine = ::chommp($logLine);
        ($logFullPath, undef) = split(" ", $logLine);
        FFP_INNER: while ($loopEval or $loopFP) {
            if ($loopEval) {
                $evalLine = ::chommp($evalLine);
                ($evalFullPath, undef) = split(" ", $evalLine);
            }
            if ($loopFP) {
                $fpLine = ::chommp($fpLine);
                ($fpFullPath, undef) = split(" ", $fpLine);
            }
            if ($logFullPath gt $evalFullPath
                    and $logFullPath gt $fpFullPath) {
                # there are missing samples AND false positves in logFile
                $evalLine = <INEVAL>;
                $fpLine = <INFP>;
                if (not defined($evalLine)) {
                    $loopEval = 0;
                    $evalFullPath = $BIGVALUE;
                    if ($::debug > 1) {
                        ::log("end of eval-file reached\n", 5);
                    }
                }
                if (not defined($fpLine)) {
                    $loopFP = 0;
                    $fpFullPath = $BIGVALUE;
                    if ($::debug > 1) {
                        ::log("end of fp-file reached\n", 5);
                    }
                }
                next FFP_INNER;
            }
            if ($logFullPath gt $evalFullPath) {
                # there are missing samples in logFile
                $evalLine = <INEVAL>;
                if (not defined($evalLine)) {
                    $loopEval = 0;
                    $evalFullPath = $BIGVALUE;
                    if ($::debug > 1) {
                        ::log("end of eval-file reached\n", 5);
                    }
                }
                next FFP_INNER;
            }
            if ($logFullPath gt $fpFullPath) {
                # there are missing false positves in logFile
                $fpLine = <INFP>;
                if (not defined($fpLine)) {
                    $loopFP = 0;
                    $fpFullPath = $BIGVALUE;
                    if ($::debug > 1) {
                        ::log("end of fp-file reached\n", 5);
                    }
                }
                next FFP_INNER;
            }
            if ($logFullPath eq $evalFullPath) {
                # correct sample found
                print(OUTEVAL $logLine, $::EOL);
                $logLine = <INLOG>;
                next FFP_OUTER;
            } elsif ($logFullPath eq $fpFullPath) {
                # correct false positve found
                print(OUTFP $logLine, $::EOL);
                $logLine = <INLOG>;
                next FFP_OUTER;
            }
            # this line is neither in evalList nor in fpList
            # it is filtered out but reported in a different file for checks
            if ($::debug) {
                ::log("line not found in lists: $logFullPath\n", 4);
                ::log("listline: $evalFullPath\n", 5);
                ::log("fpline: $fpFullPath\n", 5);
            }
            print(OUTREM $logLine, $::EOL);
            $logLine = <INLOG>;
            next FFP_OUTER;
        } # end of FFP_INNER loop
        # this line is neither in evalList nor in fpList
        # it is filtered out but reported in a different file for checks
        if ($::debug) {
            ::log("line not found in lists: $logFullPath\n", 4);
            ::log("listline: $evalFullPath\n", 5);
            ::log("fpline: $fpFullPath\n", 5);
        }
        print(OUTREM $logLine, $::EOL);
        $logLine = <INLOG>;
        next FFP_OUTER;
    } # end of FFP_OUTER loop

    close(INLOG);
    close(INEVAL);
    close(INFP);
    close(OUTEVAL);
    close(OUTFP);
    close(OUTREM);
    return 1;
}


# extract data to evaluate with one packer from logfile containing all data
sub filterPacked {
    my $tb = $_[0];     # logfile that should be filtered
    my $inLog = $_[1];      # logfile that should be filtered
    my $outAdd = $_[2];     # filename containing all lines that were added
                            # to archive list
    my $outRes = $_[3];     # filename to write all remaining lines to
    my $outPrefix = $_[4];  # prefix for filename to write all other lines to
    my $outExt = $_[5];     # extension for filename ...
    my %packPrg = %{$_[6]}; # contains list of compression programs and
                            # their archive filenames
    my $outArch = $outPrefix . $::FileExtArchiveFiles . "." . $outExt;
    my %packFile;           # contains filehandles for output files for each
                            # compression program
    my $pack;
    # set correct syntax for logfile parsing
    if ($::configSyntax eq $::OS_WINDOWS) {
        fileparse_set_fstype("MSWin32");
    } else {
        fileparse_set_fstype("Unix");
    }

    if ($::debug) {
        ::log("splitting packed logs...\n", 4);
    }

    open(INLOG, "< $inLog") or do {
        return 0;
    };
    binmode(INLOG);
    open(OUTARCH, "> $outArch") or do {
        close(INLOG);
        return 0;
    };
    binmode(OUTARCH);
    open(OUTADD, "> $outAdd") or do {
        close(INLOG);
        close(OUTARCH);
        return 0;
    };
    binmode(OUTADD);
    open(OUTRES, "> $outRes") or do {
        close(INLOG);
        close(OUTARCH);
        close(OUTADD);
        return 0;
    };
    binmode(OUTRES);

    foreach $pack (keys %packPrg) {
        my $outFile = $outPrefix . $pack . "." . $outExt;
        $packFile{$pack} = new IO::File;
        $packFile{$pack}->open("> $outFile") or do {
            close(INLOG);
            close(OUTARCH);
            close(OUTADD);
            close(OUTRES);
            my $temp;
            foreach $temp (keys %packPrg) {
                $packFile{$temp}->close;
            }
            return 0;
        };
        binmode($packFile{$pack});
    }

    # examine all entries in logfile if they match with the matchlist or not
    my $logLine;
    my $lastPath = "";
    my $archiveFound = 0;
    FILTER_PACK_LOOP: while (defined($logLine = <INLOG>)) {
        $logLine = ::chommp($logLine);
        my ($tempLoc, $tempMsg) = split(" ", $logLine, 2);
        # search for the right compression program in the sample path
        foreach $pack (keys %packPrg) {
            my $match = $packPrg{$pack}{"filename"};
            # extract the path without compression program and filename
            my $tempPath = $tempLoc;
            # only proceed if it succeeds (so we found the right packer)
            if ($tempPath =~ s/\Q$::rps$match\E(.*)$//) {
                my $content = $1;
                my $archPath = $tempPath . $::rps . $match;
                if ($archPath ne $lastPath) {
                    $archiveFound = 0;
                    $lastPath = $archPath;
                }
                if ($::debug > 1) {
                    ::log("line belongs to packer '$pack'\n", 5);
                }

                if (defined($content) and $content ne "") {
                    # handle all lines containing archive contents
                    # assemble the unpacked testbeds path by removing
                    # path separator and archive name
                    my $contentLine = $tempPath . $content . " " . $tempMsg;
                    print({$packFile{$pack}} $contentLine, $::EOL);
                    # some products report the archive as okay and only the
                    # contents as infected. These archives must also be evaluated
                    # as infected
                    if (not $archiveFound) {
                        $archiveFound = 1;
                        my $archLine = $archPath . " " . $tempMsg;
                        print(OUTARCH $archLine, $::EOL);
                        print(OUTADD $archLine, $::EOL);
                    }
                    # avoid to search for all remaining compression programs
                    next FILTER_PACK_LOOP;
                } else {
                    # handle all lines containing only an archive
                    # without contents
                    print(OUTARCH $logLine, $::EOL);
                    $archiveFound = 1;
                    next FILTER_PACK_LOOP;
                }
            }
        }
        # Each line should match at least one compression program. Therefore
        # if this code is ever reached, somewhere has an error occured.
        # All lines here are collected in a special file for error detection
        print(OUTRES $logLine, $::EOL);
    }

    close(INLOG);
    close(OUTARCH);
    close(OUTADD);
    close(OUTRES);
    foreach $pack (keys %packPrg) {
        $packFile{$pack}->close;
    }
    return 1;
}


# extract data to evaluate with one packer from logfile containing all data
sub filterPackedOk {
    my $tb = $_[0];
    my $inLog = $_[1];      # logfile that should be filtered
    my $outAdd = $_[2];     # filename containing all lines that were added
                            # to archive list
    my $outRes = $_[3];     # filename to write all remaining lines to
    my $outPrefix = $_[4];  # prefix for filename to write all other lines to
    my $outExt = $_[5];     # extension for filename ...
    my %packPrg = %{$_[6]}; # contains list of compression programs and
                            # their archive filenames
    my $outArch = $outPrefix . $::FileExtArchiveFiles . "." . $outExt;
    my %packFile;           # contains filehandles for output files for each
                            # compression program
    #my %infectedFile;       # contains a list of all compression programs and
                            # filehandles for their input files
    my $pack;
    # set correct syntax for logfile parsing
    if ($::configSyntax eq $::OS_WINDOWS) {
        fileparse_set_fstype("MSWin32");
    } else {
        fileparse_set_fstype("Unix");
    }

    if ($::debug) {
        ::log("splitting packed 'sbo' logs...\n", 5);
    }

    open(INLOG, "< $inLog") or do {
        return 0;
    };
    binmode(INLOG);
    open(OUTARCH, "> $outArch") or do {
        close(INLOG);
        return 0;
    };
    binmode(OUTARCH);
    open(OUTADD, "> $outAdd") or do {
        close(INLOG);
        close(OUTARCH);
        return 0;
    };
    binmode(OUTADD);
    open(OUTRES, "> $outRes") or do {
        close(INLOG);
        close(OUTARCH);
        close(OUTADD);
        return 0;
    };
    binmode(OUTRES);

    foreach $pack (keys %packPrg) {
        my $outFile = $outPrefix . $pack . "." . $outExt;
        $packFile{$pack} = new IO::File;
        $packFile{$pack}->open("> $outFile") or do {
            close(INLOG);
            close(OUTARCH);
            close(OUTRES);
            my $temp;
            foreach $temp (keys %packPrg) {
                $packFile{$temp}->close;
            }
            return 0;
        };
        binmode($packFile{$pack});
        #$infectedFile{$pack} = $outPrefix . $pack . "." . $::FileExtNorm;
    }

    # examine all entries in logfile if they match with the matchlist or not
    my $logLine;
    my $lastPath = "";
    my $archiveFound = 0;
    FILTER_PACK_LOOP: while (defined($logLine = <INLOG>)) {
        # 'scanned-but-ok' lines contain only the filename
        $logLine = ::chommp($logLine);
        # search for the right compression program in the sample path
        foreach $pack (keys %packPrg) {
            my $match = $packPrg{$pack}{"filename"};
            # extract the path without compression program and filename
            my $tempPath = $logLine;
            # only proceed if it succeeds (so we found the right packer)
            if ($tempPath =~ s/\Q$::rps$match\E(.*)$//) {
                my $content = $1;
                my $archPath = $tempPath . $::rps . $match;
                if ($archPath ne $lastPath) {
                    $archiveFound = 0;
                    $lastPath = $archPath;
                }
                if ($::debug > 1) {
                    ::log("line belongs to packer '$pack'\n", 5);
                }

                if (defined($content) and $content ne "") {
                    # handle all lines containing archive contents
                    # assemble the unpacked testbeds path by removing
                    # path separator and archive name
                    my $contentLine = $tempPath . $content;
                    print({$packFile{$pack}} $contentLine, $::EOL);
                    # Some products don't report the archive itself and only the
                    # contents as okay. These archives should also be evaluated
                    # as okay.
### This must be tested more thoroughly !!!
                    ### But only if there were no infected files found in
                    ### it AND all files in the archive were really scanned !!!
                    ### Since it is not possible to check these things at this time
                    ### this is not done.
                    ### As a consequence some archives might be listed in the
                    ### postscanlist although they were scanned (but no infected
                    ### file were found in it). These archive can be found in
                    ### the PROD.SBO file

                    # This test is NOT sufficient as it only checks if no
                    # infected files were found in the whole testbeds (rather
                    # than in the single archive) and no check is done whether
                    # some files in the archive are not reported
                    #if (not $archiveFound and -z $infectedFile{$pack}) {
                    #    $archiveFound = 1;
                    #    print(OUTARCH $archPath, $::EOL);
                    #    print(OUTADD $archPath, $::EOL);
                    #}
                    if (not $archiveFound) {
                        $archiveFound = 1;
                        print(OUTARCH $archPath, $::EOL);
                        print(OUTADD $archPath, $::EOL);
                    }
                    # avoid to search for all remaining compression programs
                    next FILTER_PACK_LOOP;
                } else {
                    # handle all lines containing only an archive
                    # without contents
                    print(OUTARCH $logLine, $::EOL);
                    $archiveFound = 1;
                    next FILTER_PACK_LOOP;
                }
            }
        }
        # Each line should match at least one compression program. Therefore
        # if this code is ever reached, somewhere has an error occured.
        # All lines here are collected in a special file for error detection
        print(OUTRES $logLine, $::EOL);
    }

    close(INLOG);
    close(OUTARCH);
    close(OUTADD);
    close(OUTRES);
    foreach $pack (keys %packPrg) {
        $packFile{$pack}->close;
    }
    return 1;
}


# return a true value
1;