# Export.pm
# implements 'report logs' and 'report missed' commands
#
# v0.08, 2004-04-27
# Copyright (c) Michel Messerschmidt 2003-2004
#
package Avtc::Export;
use strict;
use Exporter;
use Cwd;
use File::Basename;
use File::Copy;
use File::Path;

@Avtc::Export::ISA = qw(Exporter);
@Avtc::Export::EXPORT = qw(exportLogs exportMissed);

####################### implemented commands ###############################

sub exportLogs {
    my $ret = 0;
    ::log("Working on test: $::configTestName\n");
    # products can be selected for this function while all testbeds and
    # systems must be processed for each product
    LOOP_PROD: foreach my $prod (sort keys %::prodList) {
        # this is set to 1 ("true"), if at least one logfile for this
        # product is not finished yet
        my $preLogs = 0;
        my $tempDir = $::TempBase .$::sep. "LOGS_" . $prod;
        # check existance of tempDir and create it if nessecary
        ::makeDir($tempDir . $::sep) or do {
            ::log("ERROR: creating directory $tempDir\n", 1);
            $ret = 1;
        };
        # change to tempDir, so all following IO operates on current dir
        chdir($tempDir) or do {
            ::log("ERROR: chdir $tempDir failed:  $!\n", 1);
            ::log("Export of $prod logfiles failed !\n\n", 4);
            $ret = 1;
            next LOOP_PROD;
        };
        if ($::debug) {
            ::log("Current dir set to ". cwd() ."\n", 1);
        }

        foreach my $os (sort (keys %::osListAll)) {
            LOOP_TB: foreach my $tb (sort keys %::listsAll) {
                # now we can define the absolute paths, we need
                my $EvalDir = $::EvalBase . $::sep . $os . $::sep . $tb;
                my $LogDir = $::LogBase . $::sep . $os . $::sep . $tb;

                # check for not tested combinations
                foreach (@::noTestListAll) {
                    if ($_ eq "$os $tb $prod") {
                        if ($::debug) {
                            ::log("not tested: $os - $tb - $prod\n", 1);
                        }
                        next LOOP_TB;
                    }
                }
                ::log("processing $prod - $tb - $os\n", 1);
                # define name of product related files
                my $statFile = $EvalDir . $::sep . $prod . $::FileExtStatus;
                my $ful = $prod . "." . $::FileExtReport;
                # filenames for postscan logs
                my %ful = ();
                my $archive = $EvalDir . $::sep . $prod . "." . $::ZipExt;
                my $status = ::getStatus($statFile);

                if ($status eq "") {
                    ::log("ERROR getting status for $os - $tb - $prod. ".
                          "Export canceled\n", 2);
                    $ret = 1;
                    next LOOP_PROD;
                } elsif ($status =~ /^NO/i or $status =~ /^-S1/i) {
                    if ($::debug) {
                        ::log("status checked, no logfile for $prod on ".
                              "testbed $tb\n", 2);
                    }
                    next LOOP_TB;
                } elsif ($status !~ /^OK/i) {
                    $preLogs = 1;
                    ::log("WARNING: test not finished for $os - $tb - $prod.",2);
                    ::log("         Creating preliminary log archive\n", 2);
                }
                # else: 'ok' status means test is finished - go ahead

                # get number of scans (= number of logfiles)
                my $scanNr;
                if ($status =~ /^-S(\d)/i) {
                    $scanNr = $1 - 1;
                    if ($scanNr < 1) {
                        $scanNr = 1;
                    }
                } elsif ($status =~ /^(OK )?S(\d)E?/i) {
                    $scanNr = $2;
                } else {
                    # just a default value: no postscan logs
                    $scanNr = 1;
                }
                if ($::debug > 2) {
                    ::log("number of scan logfiles: '$scanNr'\n", 2);
                }

                # handle first scan
                if ($scanNr == 1) {
                    # if there is only one logfiles it has still the name 'XXX.ful'
                    # and wasn't renamed to 'XXXs1.ful' (this is done at the first postscan)
                    $ful{"1"} = $prod . "." . $::FileExtReport;
                } else {
                    $ful{"1"} = $prod . "S1." . $::FileExtReport;
                }
                if ($::debug > 2) {
                    ::log("added log '". $ful{"1"} . "' for scan 1\n");
                }
                # postscans have numbers one lower than the scanNr:
                # 1. postscan (XXXn1.ful) is the 2. scan
                foreach (my $i = 1; $i < $scanNr; $i++) {
                    # use the scanNr not the postscanNr for the hash key
                    my $j = $i + 1;
                    $ful{$j} = $prod . "N" . $i . "." . $::FileExtReport;
                    if ($::debug > 2) {
                        ::log("added log '". $ful{$j} . "' for scan $j (postscan $i)\n");
                    }
                }

                # get logfile from evaluation data archive
                if ($::configOs eq $::OS_WINDOWS) {
                    $archive = ::winPath($archive);
                }
                my $fulList = join(" ", values(%ful));
                my $extCall = $::UnzipProg . " " . $::UnzipOpt . " " .
                   $archive . " " . $fulList;

                if ($::debug) {
                    ::log("\nlist of logfiles to extract: '$fulList'\n", 2);
                    ::log("calling '$extCall'\n", 2);
                }
                my $error = system($extCall);
                if ($::debug) {
                    ::log("...external program terminated with ".
                          "exitcode $error\n", 2);
                }
                if ($error) {
                    ::log("ERROR calling ".$::UnzipProg.": exit code=".
                          $error.", error=$!\n", 2);
                    ::log("Export of $prod logfiles failed !\n\n", 4);
                    $ret = 1;
                    next LOOP_PROD;
                }

                # rename logfile (otherwise it would be overwritten with the
                # next logfile)
                my %dest = ();
                $dest{"1"} = $os . "_" . $tb . "." . $::FileExtLog;
                foreach (my $i = 1; $i < $scanNr; $i++) {
                    # once again: use the scanNr not the postscanNr for the hash key
                    my $j = $i + 1;
                    $dest{"$j"} = $os . "_" . $tb . "-post" . $i . "." . $::FileExtLog;
                }
                foreach my $nr (sort keys %ful) {
                    my $s = $ful{$nr};
                    my $d = $dest{$nr};
                    move($s, $d) or do {
                        ::log("ERROR moving '$s' to '$d': $!\n", 2);
                        ::log("Export of $prod logfiles failed !\n\n", 4);
                        $ret = 1;
                        next LOOP_PROD;
                    };
                }

            }  #end of foreach LOOP_TB
        }  #end of foreach os

        # if no logfile was extracted there's nothing left to do for this
        # product
        if (::fileCount($tempDir) < 1) {
            next LOOP_PROD;
        }
        # create the 'readme.txt' file
        my $template = $::EvalRoot . $::sep . $::configDirNames{"template"} .
                       $::sep . $::FileExportLogReadme;
        if ($preLogs) {
            # prepend an additional warning to the readme file for
            # preliminary results
            my $preTxt = $::EvalRoot .$::sep. $::configDirNames{"template"} .
                         $::sep . $::FileExportLogPreReadme;
            copy($preTxt, "readme.txt") or do {
                ::log("ERROR copying $preTxt to current directory: ".
                      "$!\n", 2);
                ::log("Export of $prod logfiles failed !\n\n", 4);
                $ret = 1;
                next LOOP_PROD;
            };
            # append template readme file
            ::appendFile("readme.txt", $template)  or do {
                ::log("ERROR appending $template to ".
                      "readme.txt: $!\n", 2);
                ::log("Export of $prod logfiles failed !\n\n", 4);
                $ret = 1;
                next LOOP_PROD;
            };
        } else {
            # copy template readme file to current directory
            copy($template, "readme.txt") or do {
                ::log("ERROR copying $template to current directory: ".
                      "$!\n", 2);
                ::log("Export of $prod logfiles failed !\n\n", 4);
                $ret = 1;
                next LOOP_PROD;
            };
        }
        # create archive with all logfiles for this product
        # this is created in the current dir and moved to the server later on
        my $logarchive = $prod . "_logs" . "." . $::ZipExt;
        my $wildcard = "*.*";
        if ($::configOs eq $::OS_UNIX) {
            $wildcard = "*";
        }

        my $extCall = $::ZipProg . " " . $::ZipOpt . " " .
           $logarchive . " " . $wildcard;

        if ($::debug) {
            ::log("calling $extCall\n", 2);
        }
        my $error = system($extCall);
        if ($::debug) {
            ::log("...external program terminated with ".
                  "exitcode $error\n", 2);
        }
        if ($error) {
            ::log("ERROR calling ".$::ZipProg.": exit code=".
                  $error.", error=$!\n", 2);
            ::log("Export of $prod logfiles failed !\n\n", 4);
            $ret = 1;
            next LOOP_PROD;
        }

        # move logfile archive to the server
        my $dest = $::EvalBase . $::sep . $::configDirNames{"report"} .
                   $::sep . $logarchive;
        copy($logarchive, $dest) or do {
            ::log("ERROR copying '".$logarchive."' to '".$dest."': $!\n", 2);
            ::log("Export of $prod logfiles failed !\n\n", 4);
            $ret = 1;
            next LOOP_PROD;
        };
    }  #end of foreach LOOP_PROD

    # delete all temporary files that were created here
    chdir($::TempBase) or do {
        ::log("ERROR: chdir $::TempBase failed:  $!\n", 1);
        ::log("Deletion of temporary files failed !\n", 4);
        return 1;
    };
    if ($::debug) {
        ::log("Current dir set to ". cwd() ."\n", 1);
    }
    ::deleteFiles($::TempBase, '^LOGS_', 1);
    return $ret;
}


sub exportMissed {
    my $ret = 0;
    ::log("Working on test: $::configTestName\n");
    # products can be selected for this function while all testbeds and
    # systems must be processed for each product
    LOOP_PROD: foreach my $prod (sort keys %::prodList) {
        my $archDir = $::TempBase .$::sep. "MISS_" . $prod;
        LOOP_OS: foreach my $os (sort (keys %::osListAll)) {
            my $tempDir = $archDir . $::sep . $os;
            # check existance of tempDir and create it if nessecary
            ::makeDir($tempDir . $::sep) or do {
                ::log("ERROR: creating directory $tempDir\n", 1);
                $ret = 1;
            };
            # change to tempDir, so all following IO operates on current dir
            chdir($tempDir) or do {
                ::log("ERROR: chdir $tempDir failed:  $!\n", 1);
                ::log("Export of 'missed samples' list for $prod failed !\n\n", 4);
                $ret = 1;
                next LOOP_PROD;
            };
            if ($::debug) {
                ::log("Current dir set to ". cwd() ."\n", 1);
            }

            LOOP_TB: foreach my $tb (sort keys %::listsAll) {
                if (exists $::listsAll{$tb}->{"packed"}) {
                    if ($::debug) {
                        ::log("missed samples are not exported for packed ".
                              "testbeds ($tb)\n", 1);
                    }
                    next LOOP_TB;
                }

                # now we can define the absolute paths, we need
                my $EvalDir = $::EvalBase . $::sep . $os . $::sep . $tb;
                my $LogDir = $::LogBase . $::sep . $os . $::sep . $tb;

                # check for not tested combinations
                foreach (@::noTestListAll) {
                    if ($_ eq "$os $tb $prod") {
                        if ($::debug) {
                            ::log("not tested: $os - $tb - $prod\n", 1);
                        }
                        next LOOP_TB;
                    }
                }
                ::log("processing $prod - $tb - $os\n", 1);
                # define name of product related files
                my $statFile = $EvalDir . $::sep . $prod . $::FileExtStatus;
                my $mis = $prod . "." . $::FileExtMissedFiles;
                my $dfp = $prod . "." . $::FileExtDetFPFiles;
                my $archive = $EvalDir . $::sep . $prod . "." . $::ZipExt;
                my $status = ::getStatus($statFile);

                if ($status eq "") {
                    ::log("ERROR getting status for $os - $tb - $prod. ".
                          "Export canceled\n", 2);
                    $ret = 1;
                    next LOOP_PROD;
                } elsif ($status =~ /^NO/i) {
                    # don't check for status '-S1' because only finished
                    # logfiles can be processed here
                    if ($::debug) {
                        ::log("status checked, no logfile for $prod on ".
                              "testbed $tb\n", 2);
                    }
                    next LOOP_TB;
                } elsif ($status !~ /^OK/i) {
                    ::log("ERROR: test not finished for $os - $tb - $prod. ".
                          "Export canceled\n", 2);
                    $ret = 1;
                    next LOOP_PROD;
                }
                # else: 'ok' status means test is finished - go ahead

                # get 'missed samples' list from evaluation data archive
                if ($::configOs eq $::OS_WINDOWS) {
                    $archive = ::winPath($archive);
                }
                my $extCall = $::UnzipProg . " " . $::UnzipOpt . " " .
                   $archive . " " . $mis;

                if ($::debug) {
                    ::log("calling $extCall\n", 2);
                }
                my $error = system($extCall);
                if ($::debug) {
                    ::log("...external program terminated with ".
                          "exitcode $error\n", 2);
                }
                if ($error) {
                    ::log("ERROR calling ".$::UnzipProg.": exit code=".
                          $error.", error=$!\n", 2);
                    ::log("Export of 'missed samples' list for $prod failed !\n\n", 4);
                    $ret = 1;
                    next LOOP_PROD;
                }

                # rename list (otherwise it would be overwritten with the
                # next list)
                my $dest = $tb . "." . $::FileExtMissedFiles;
                my $destFp = $tb . "." . $::FileExtDetFPFiles;
                move($mis, $dest) or do {
                    ::log("ERROR moving '".$mis."' to '".$dest."': $!\n", 2);
                    ::log("Export of 'missed samples' list for $prod failed !\n\n", 4);
                    $ret = 1;
                    next LOOP_PROD;
                };

                # get 'detected false positives' list from evaluation data archive
                # but only if this testbed contains false positives :-)
                if (exists $::listsAll{$tb}->{"fp"}) {
                    $extCall = $::UnzipProg . " " . $::UnzipOpt . " " .
                       $archive . " " . $dfp;

                    if ($::debug) {
                        ::log("calling $extCall\n", 2);
                    }
                    my $error = system($extCall);
                    if ($::debug) {
                        ::log("...external program terminated with ".
                              "exitcode $error\n", 2);
                    }
                    if ($error) {
                        ::log("ERROR calling ".$::UnzipProg.": exit code=".
                              $error.", error=$!\n", 2);
                        ::log("Export of 'missed samples' list for $prod failed !\n\n", 4);
                        $ret = 1;
                        next LOOP_PROD;
                    }

                    move($dfp, $destFp) or do {
                        ::log("ERROR moving '".$dfp."' to '".$destFp."': $!\n", 2);
                        ::log("Export of 'missed samples' list for $prod failed !\n\n", 4);
                        $ret = 1;
                        next LOOP_PROD;
                    };

#                    ::appendFile($dest, $dfp) or do {
#                        ::log("ERROR appending $dfp to $dest: $!\n", 2);
#                        ::log("Export of 'missed samples' list for $prod failed !\n\n", 4);
#                        $ret = 1;
#                        next LOOP_PROD;
#                    };
#                    my $deleted = unlink($dfp);
#                    if (not $deleted) {
#                        ::log("ERROR deleting file'$dfp': $!\n", 2);
#                        ::log("Export of 'missed samples' list for $prod failed !\n\n", 4);
#                        $ret = 1;
#                        next LOOP_PROD;
#                    }
                }

            }  #end of foreach LOOP_TB
        }  #end of foreach os

        # change to tempDir, so all following IO operates on current dir
        chdir($archDir) or do {
            ::log("ERROR: chdir $archDir failed:  $!\n", 1);
            ::log("Export of 'missed samples' list for $prod failed !\n\n", 4);
            $ret = 1;
            next LOOP_PROD;
        };
        if ($::debug) {
            ::log("Current dir set to ". cwd() ."\n", 1);
        }
        # create archive with all missed samples lists for this product
        # this is created in the current dir and moved to the server later on
        my $misArch = $prod . "_miss" . "." . $::ZipExt;
        my $wildcard = "*.*";
        if ($::configOs eq $::OS_UNIX) {
            $wildcard = "*";
        }

        my $extCall = $::ZipProg . " " . $::ZipOpt . " " .
           $misArch . " " . $wildcard;

        if ($::debug) {
            ::log("calling $extCall\n", 2);
        }
        my $error = system($extCall);
        if ($::debug) {
            ::log("...external program terminated with ".
                  "exitcode $error\n", 2);
        }
        if ($error) {
            ::log("ERROR calling ".$::ZipProg.": exit code=".
                  $error.", error=$!\n", 2);
            ::log("Export of 'missed samples' list for $prod failed !\n\n", 4);
            $ret = 1;
            next LOOP_PROD;
        }

        # move logfile archive to the server
        my $dest = $::EvalBase . $::sep . $::configDirNames{"report"} .
                   $::sep . $misArch;
        copy($misArch, $dest) or do {
            ::log("ERROR copying '".$misArch."' to '".$dest."': $!\n", 2);
            ::log("Export of 'missed samples' list for $prod failed !\n\n", 4);
            $ret = 1;
            next LOOP_PROD;
        };
    }  #end of foreach LOOP_PROD

    # delete all temporary files that were created here
    chdir($::TempBase) or do {
        ::log("ERROR: chdir $::TempBase failed:  $!\n", 1);
        ::log("Deletion of temporary files failed !\n", 4);
        return 1;
    };
    if ($::debug) {
        ::log("Current dir set to ". cwd() ."\n", 1);
    }
    ::deleteFiles($::TempBase, '^MISS_', 1);
    return $ret;
}


# return a true value
1;