# generic scanner key for multi-line logfiles
# version 6.00 - 2004-04-05
# created by Michel Messerschmidt
# IMPORTANT: All lines directly following a line beginning with "### EXAMPLE:"
# MUST be adjusted for the actual scanner !!!

use strict;
use File::Basename;

# get data from calling program (see documentation in 'scanner-key.txt')
# The hash %::KeyAPI is created in Eval.pm in the 'normalizeLog' subroutine
# (look there if the documentation is outdated).
# PLEASE use $EOL for line-ends instead of '\n'.
# Additionally all file MUST be opened in binary mode (see below).
#
# Get some constants
my $EOL = $::KeyAPI{"constEOL"};
my $OS_UNIX = $::KeyAPI{"constUNIX"};
my $OS_WIN = $::KeyAPI{"constWINDOWS"};
my $VirusNameUI = $::KeyAPI{"nameUI"};
# Get debug level
my $debug = $::KeyAPI{"debug"};
# Get a reference to the logging function (may be called with: &$logRef();
my $logRef = $::KeyAPI{"logRef"};
# Get filenames
my $inLog = $::KeyAPI{"inLog"};
my $outNorm = $::KeyAPI{"outNorm"};
my $outOkay = $::KeyAPI{"outOkay"};
my $outRest = $::KeyAPI{"outRest"};
my $outEval = $::KeyAPI{"outEval"};
my $outEvalRed = $::KeyAPI{"outEvalRedundant"};
my $fileCut = $::KeyAPI{"outCut"};
my $fileControl = $::KeyAPI{"outControl"};
my $fileCorrupt = $::KeyAPI{"outCorrupt"};
# Get some global directories
my $boot = $::KeyAPI{"boot"};
my $tempDir = $::KeyAPI{"tempDir"};
my $binDir = $::KeyAPI{"binDir"};
# Get configuration of the current evaluation
my $os = $::KeyAPI{"os"};
my $tb = $::KeyAPI{"testbed"};
my $prod = $::KeyAPI{"product"};
my $syntax = $::KeyAPI{"syntax"};
my $pathUnix = $::KeyAPI{"pathUnix"};
my $pathWin = $::KeyAPI{"pathWindows"};
my $sep = $::KeyAPI{"sep"};
my $seprun = $::KeyAPI{"sepRun"};
my $tbRoot = $::KeyAPI{"testbedRoot"};
my $tbID = $::KeyAPI{"testbedID"};
# A regular expression to match exactly all sample filenames
my $sampleRE = $::KeyAPI{"samplenameRE"};
# A regular expression to match exactly all archive filenames in
# packed testbeds. This is empty for all other testbeds.
my $packerRE = "";
# A regular expression that combine $sampleRE and $fileRE to match
# all files that must be evaluated
# WARNING: Although $fileRE is the same as $sampleRE on most testbeds,
# it is extended by $packerRE on packed testbeds (see below).
# Therefore it MUST always be used in parentheses !!!
# Example: $inLine =~ /^x:\\foo\\bar\\($fileRE)/i
my $fileRE = $::KeyAPI{"samplenameRE"};
# Contains the archive filenames for all compression programs
my %packPrg;
# boolean flag that is true for packed testbeds
my $packed = $::KeyAPI{"packed"};
if ($packed) {
    %packPrg = %{$::KeyAPI{"packRef"}};
    # Create the regular expression to match all archive filenames
    # in packed testbeds.
    # WARNING: Always use this in parentheses or it will NOT work !!!
    # Example: $inLine =~ /^x:\\foo\\bar\\($packerRE)/i
    foreach my $pack (keys %packPrg) {
        $packerRE .= $packPrg{$pack}{"filename"} . "|";
    }
    chop($packerRE);
    # $fileRE is extended to match match sample AND archive filenames
    $fileRE .= "|" . $packerRE;
}
if ($debug > 1) {
    &$logRef("regular expression for sample filenames:  '".$sampleRE."'\n", 4);
    &$logRef("regular expression for archive filenames: '".$packerRE."'\n", 4);
    &$logRef("regular expression for evaluated files:   '".$fileRE."'\n", 4);
}

# All data about samples is collected in this hash:
# each key identifies one sample (with full path)
# all "infected" reports are collected in the sub-hash
# %{$data{$samplePath}->{"infected"}}
# with key = virusname and value = array of all lines
# all lines with "okay" reports are collected in the sub-array
# @{$data{$samplePath}->{"okay"}}
# The array @{$data{$samplePath}->{"redundant"}} contains all logfile
# lines related to a sample but not used for evaluation
my %data = ();
# All line belonging to NO sample are collected here
my @linesRemain = ();
# All samples reported as "corrupt" are *additionally* collected here
my @corruptSamples = ();
# Count all files reported as corrupt
my $corruptNr = 0;
# Number of detected files reported by scanner
my $sumInfected = 0;
# Number of scanned files reported by scanner
my $sumScanned = 0;
# Counts number of lines in input file
my $lineNr = 0;
# Contains current line from input file
my $inLine = "";
# full path (with filename) of sample
my $samplePath = "";
# reported virusname of infected sample
my $virusName = "";
# The function 'detectUnreliable' may override the infection status with
# this flag (to prevent counting Eicar-test-files as viruses)
my $testFile = 0;

# set correct syntax for logfile parsing
if ($syntax eq $OS_WIN) {
    fileparse_set_fstype("MSWin32");
} else {
    fileparse_set_fstype("Unix");
}


open (IN, "< $inLog") or do {
    &$logRef("ERROR opening $inLog: $!\n", 2);
    return 0;
};
# Some systems (dos, windows !!!) distinguish between binary and text files.
# To disable implicit text conversion on these systems (LF <-> CR+LF) all
# files MUST be opened in binary mode.
binmode(IN);
open (INFECTED, "> $outNorm") or do {
    &$logRef("ERROR opening $outNorm: $!\n", 2);
    close(IN);
    return 0;
};
binmode(INFECTED);
open (OKAY, "> $outOkay") or do {
    &$logRef("ERROR opening $outOkay: $!\n", 2);
    close(IN);
    close(INFECTED);
    return 0;
};
binmode(OKAY);
open (REMAINS, "> $outRest") or do {
    &$logRef("ERROR opening $outRest: $!\n", 2);
    close(IN);
    close(INFECTED);
    close(OKAY);
    return 0;
} ;
binmode(REMAINS);
open (EVAL, "> $outEval") or do {
    &$logRef("ERROR opening $outEval: $!\n", 2);
    close(IN);
    close(INFECTED);
    close(OKAY);
    close(REMAINS);
    return 0;
};
binmode(EVAL);
open (EVALRED, "> $outEvalRed") or do {
    &$logRef("ERROR opening $outEvalRed: $!\n", 2);
    close(IN);
    close(INFECTED);
    close(OKAY);
    close(REMAINS);
    close(EVAL);
    return 0;
};
binmode(EVALRED);
open (CUT, "> $fileCut") or do {
    &$logRef("ERROR opening $fileCut: $!\n", 2);
    close(IN);
    close(INFECTED);
    close(OKAY);
    close(REMAINS);
    close(EVAL);
    close(EVALRED);
    return 0;
};
binmode(CUT);
open (CONTROL, "> $fileControl") or do {
    &$logRef("ERROR opening $fileControl: $!\n", 2);
    close(IN);
    close(INFECTED);
    close(OKAY);
    close(REMAINS);
    close(EVAL);
    close(EVALRED);
    close(CUT);
    return 0;
};
binmode(CONTROL);
open (CORRUPT, "> $fileCorrupt") or do {
    &$logRef("ERROR opening $fileCorrupt: $!\n", 2);
    close(IN);
    close(INFECTED);
    close(OKAY);
    close(REMAINS);
    close(EVAL);
    close(EVALRED);
    close(CUT);
    close(CONTROL);
    return 0;
};
binmode(CORRUPT);


&$logRef(" ", 3);
LINE_LOOP: while (defined($inLine = <IN>)) {
    # Count all read lines (needed for statistics)
    $lineNr++;
    # Give some life sign in regular intervalls
    # As this uses the internal log function, the GUI should be updated too
    if ($lineNr % 5000 == 0) {
        &$logRef(".", 0);
    }
    # Cut of all control characters (like LF, CR LF, ...) from the
    # beginning and end of the line
    $inLine = ::Avtc::Tools::chommp($inLine);
    # Is the evaluation system using windows or unix syntax ?
    # convert all input lines accordingly
    # Only the beginning of the path is modified here,
    # slash <-> backslash conversion must be done later (because it can
    # be useful to use the original to determine certain filename parts)
    if ($syntax eq $OS_WIN) {
        $inLine =~ s-\Q$pathUnix\E\/-$pathWin\\-;
        # Make the complete line uppercase (don't do this with unix syntax!)
        $inLine = uc($inLine);
    } else {
        $inLine =~ s-\Q$pathWin\E\\-$pathUnix\/-;
    }

    my @field = split(" ", $inLine);
    # There are two different methods to extract the interesting parts:
    # - with 'split', if the interesting data is *always* separated by spaces
    #   (see previous line)
    # - with regular expressions:
    #   . = matches any character (but only one character)
    #   * = the previous character (or expression in parentheses) may appear
    #       0, 1 or more times in sequence
    #   + = the previous character (or expression in parentheses) may appear
    #       1 or more times in sequence
    #   ^ = matches the beginning of the line
    #   $ = matches the end of the line or is may be interpreted as a variable
    #       name (depending on context:
    #       /^abc$/ is definitly an end of line but in
    #       /^hello $name/ $name is taken as a variable name
    #       Variable name are substituted by their contents *before*
    #       interpretation of the expression, so they may contain special
    #       characters:
    #       If $name = "world", then /^hello $name/ matches all lines
    #       beginning with "hello world.
    #       If $name = ".*", then /^hello $name/ matches all lines
    #       beginning with "hello " followed by any (or no) characters.
    #   \ = the next character is taken literally, so it loses all
    #       special meaning it have have in regular expressions
    #       (e.g. use \\ to match a single backslash)
    #  \Q
    #  \E = everything between \Q and \E is taken literally
    #  () = save the substring between the parentheses in a positional
    #       variable (first parentheses in $1, second in $2, ...)
    #  [abc123] = matches on 1 character, that is contained in the brackets
    #  [^abc123] = matches on 1 character, NOT contained in the brackets
    #  \d = a numerical (decimal) character: 0,1,2,3,4,5,6,7,8,9
    #  \D = a non-numerical character: all characters not matched by \d
    #  \s = a whitespace character: ' ' (space), '\t' (tab), ...
    #  \S = a non-whitespace character: everything that is not matched by \s
    # In most cases it's better to use '\s' than ' ' to match a whitespace
    # because some reports contain tabs (ASCII 09) instead of spaces (ASCII 32)
    # that are NOT matched by ' ', while '\s' matches any whitespace character

    # Only lines that contain the current testbed root directory
    # and a valid filename are evaluated

    # Example taken from ANT:
    # This only matches lines that begin with the testbed directory and
    # contain a valid samplename
    # WARNING: this works for ANT but for most products there are lines
    #          without a samplename in packed testbeds
### EXAMPLE:
    if ($inLine =~ /^(\Q$tbRoot$sep\E.*($fileRE).*)$/) {
       # Extract the complete path and filename
        $samplePath = normalizeSample($1);
        # This is a redundant line because infected samples have another
        # (more important) line and no extra code is required for
        # okay samples (if there are only redundant lines for a sample,
        # the sample is treated as okay)
        push @{$data{$samplePath}->{"redundant"}}, $inLine;
    }
    # Every infected message begins with 'ALERT:'
### EXAMPLE:
    elsif ($inLine =~ /^ALERT:\s\[(\S+).*virus\]\s(\Q$tbRoot$sep\E.*($fileRE).*?)\s+<<</i) {
        # Extract the complete path and filename
        $samplePath = $2;
        # Extract the name with which this infection was identified.
        # Sometimes the virus/malware is not identified exactly.
        $virusName = detectUnreliable($1);
        # Check path for correctness and modify it if not
        $samplePath = normalizeSample($samplePath);
        if (not $testFile) {
            # Don't count this as infected if the $testFile flag
            # was set by the 'detectUnreliable' function
            push @{$data{$samplePath}->{"infected"}->{$virusName}}, $inLine;
        } else {
            $testFile = 0;
            push @{$data{$samplePath}->{"okay"}}, $inLine;
        }
    }
    # Now sort out some lines as redundant that are known to contain no useful
    # information. This is done to reduce the size of the 'remaining lines'
    # file (to make it more readable).
### EXAMPLE:
    elsif ($inLine =~ /^\Q$tbRoot$sep\E/) {
        # All lines that begin with the testbed root directory but don't
        # contain a valid filename are sorted out here.
        
        # These lines are collected under the last found $samplePath.
        # But that is just a technical requirement that is irrelevant 
        # for the test evaluation.
        push @{$data{$samplePath}->{"redundant"}}, $inLine;
    }
### EXAMPLE:
    elsif ($inLine =~ /^\s*$/ or
           $inLine =~ /Date:.+Time:.+Size:/i or
           $inLine =~ /Datum:/i) {
        # sort out empty lines, lines with only whitespace and
        # lines that contain time stamps
        push @{$data{$samplePath}->{"redundant"}}, $inLine;
    }
    else {
        # All other lines contain no information about testbed files
        # and are not evaluated any further.
        push(@linesRemain, $inLine);

        # The only interesting things here are the statistics that some
        # products write to the reportfile.
### EXAMPLE:
        if ($inLine =~ /^\s+files:\s+(\d+)$/i) {
            $sumScanned = $1;
        }
### EXAMPLE:
        if ($inLine =~ /^\s+alerts:\s+(\d+)$/i) {
            $sumInfected += $1;
        }
### EXAMPLE:
        if ($inLine =~ /^\s+Dateien:\s+(\d+)$/i) {
            $sumScanned = $1;
        }
### EXAMPLE:
        if ($inLine =~ /^\s+Infiziert:\s+(\d+)$/i) {
            $sumInfected += $1;
        }
    }
} # end of LINE_LOOP


# write all data
writeData();

close (IN);
close (INFECTED);
close (OKAY);
close (REMAINS);
close (EVAL);
close (EVALRED);
close (CUT);
close (CONTROL);
close (CORRUPT);

# return success
undef &writeData;
undef &normalizeSample;
undef &modifyPacked;
undef &modifyPackedCore;
undef &correctPath;
undef &detectUnreliable;
return 1;


############################### subroutines ################################

# write all data after the complete logfiles has beeen read
sub writeData {
    # Write all "redundant" lines that occured before the first line
    # with a sample filename (these are collected with the
    # empty string "" as samplename).
    foreach my $line (@{$data{""}->{"redundant"}}) {
        print(EVAL $line, $EOL);
        print(EVALRED $line, $EOL);
    }
    # Writing in sorted order makes some files more readable.
    # However it is not required for evaluation (important files are
    # re-sorted later) and can be disabled here to reduce runtime
    #foreach my $sample (keys %data) {
    foreach my $sample (sort keys %data) {
        # The empty string "" may occur as samplename (see above)
        # and must be skipped here
        next if ($sample eq "");
        # write all infected messages
        if (exists($data{$sample}->{"infected"})) {
            # this takes into account all virusnames reported for one sample
            foreach my $name (keys %{$data{$sample}->{"infected"}}) {
                my @lines = @{$data{$sample}->{"infected"}->{$name}};
                foreach my $line (@lines) {
                    # This write a line for every original line in the report
                    # These are later evaluated as "multiple scanned" lines.
                    # (This seeems to be not nessacary but simplifies error
                    # checking.)
                    print(INFECTED $sample, " ", $name, $EOL);
                    print(EVAL $line, $EOL);
                }
            }
        }
        # write all okay messages
        if (exists($data{$sample}->{"okay"})) {
            foreach my $line (@{$data{$sample}->{"okay"}}) {
                # Same as for infected messages:
                # We preserve the original messages although they are
                # redundant
                print(OKAY $sample, $EOL);
                print(EVAL $line, $EOL);
            }
        }

        # if there are no messages at all (neither infected nor okay)
        # the sample is treated as if it it reported as okay
        if (not exists($data{$sample}->{"infected"})
            and not exists($data{$sample}->{"okay"})) {
            print(OKAY $sample, $EOL);
            # remove first line from redundant lines (because it
            # belongs to this okay report and is not redundant)
            my $mainLine = shift @{$data{$sample}->{"redundant"}};
            print(EVAL $mainLine, $EOL);
            if ($debug > 1) {
                &$logRef("found implicit okay sample: '".$mainLine."'\n\n", 4);
            }
        }

        # These are all lines that are not directly related to the evaluation
        # but have to do with the current sample (for example messages like
        # "packed with")
        foreach my $line (@{$data{$sample}->{"redundant"}}) {
            print(EVAL $line, $EOL);
            print(EVALRED $line, $EOL);
        }
    }

    # write all lines that were not used for sample evaluation
    # These are basically the report header and footer lines
    foreach my $line (@linesRemain) {
        print(REMAINS $line, $EOL);
    }

    # write the list of corrupt samples
    foreach my $sample (@corruptSamples) {
        print(CORRUPT $sample, $EOL);
    }

    # output statistics to control file
    print(CONTROL "---- data extracted by scanner key (compare by hand with data below) ----", $EOL);
    printf(CONTROL "     scanned files reported by product: %7d  (= lines in %s.S_F ?)%s", $sumScanned, $prod, $EOL);
    printf(CONTROL "    infected files reported by product: %7d  (= lines in %s.EV1 ?)%s", $sumInfected, $prod, $EOL);
    if ($corruptNr) {
        printf(CONTROL "                   corrupt files found: %7d%s", $corruptNr, $EOL);
    }
    printf(CONTROL "             lines read by scanner key: %7d  (= lines in %s.FUL ?)%s", $lineNr, $prod, $EOL);
    print(CONTROL "-------------------------------------------------------------------------", $EOL);
}


# make all required modifications to the samplepath
sub normalizeSample {
    my $path = $_[0];
    if ($debug > 1) {
        &$logRef("normalizing samplepath: '".$path."'\n", 4);
    }
    if ($syntax eq $OS_WIN) {
        # convert all '/' to '\'
        $path =~ s-\/-\\-g;
    } else {
        # convert all '\' to '/'
        $path =~ s-\\-\/-g;
    }
    # special handling of packed testbeds (path must be modified sometimes)
    if ($packed) {
        $path = modifyPacked($path);
    } else {
        $path = correctPath($path);
    }
    if ($debug > 1) {
        &$logRef("new samplepath: '".$path."'\n\n", 4);
    }
    return $path;
}


### There are different methods for different scanners to modify and correct
### paths. Choose the one that best fits the report format and delete all
### other functions with the same name!

# Modify path for packed testbeds
# Outputformat is: path to sample/archive + archive name + filename
# There must be no additional directory parts between archiv-name and filename
# This one cuts everything between the archive filename and the last filename
# (identified by the last slash or backslash)
sub modifyPacked {
    my $oldPath = $_[0];
    if ($debug > 2) {
        &$logRef("modify packed sample: '".$oldPath."'\n", 4);
    }
    # look for all compression programs
    PACK_LOOP: foreach my $pack (keys %packPrg) {
        # TAR is a special case for packed testbeds of type Q as it may also
        # be embedded in other archives types
        if ($packed eq "Q" and $pack eq "TAR") { next }
        my $newPath = modifyPackedCore($pack, $oldPath);
        # "-1" means that this packers' archive filename didn't match
        # every other return value (= the modified path) indicates a match
        # Therefore the modification is complete for all other cases than "-1"
        return $newPath unless ($newPath eq "-1");
    }
    if ($packed eq "Q") {
        # special case: TAR must always be checked as last on packed type Q,
        # as it is embedded in many other archives
        my $newPath = modifyPackedCore("TAR", $oldPath);
        return $newPath unless ($newPath eq "-1");
    }
    # return unmodified parth if everything else fails
    # (this is just for safety and should normally not occur)
    return $oldPath;
}


# This one is used for products, that append the archive contents
# with a path separator to the archive filename.
sub modifyPackedCore {
    my $pack = $_[0];
    my $oldPath = $_[1];
    my $newPath;
    my $packFile = $packPrg{$pack}{"filename"};
    # only proceed for the correct packer
    # "-1" is used to signal a non-match, as it never can be a valid
    # absolute filepath
    return "-1" unless ($oldPath =~ /^(.+)\Q$sep$packFile\E(\Q$sep\E(.*))?$/);
    my $dir = $1;
    my $content = $3;
    while ($dir =~ /^(.+)\Q$sep$packFile\E$/) {
        $dir =~ s/^(.+)\Q$sep$packFile\E$/$1/;
    }

    if ($debug > 1) {
        &$logRef("found '".$packFile."'  (packer ".$pack.")\n", 5);
        &$logRef("dir='".$dir."'  -  content='".$content."'\n", 5);
    }
    # assemble new path
    $newPath = $dir . $sep . $packFile;

    # only process contents with a valid sample filename
    # (to avoid any garbage appended to the filename)
    if (defined($content) and $content =~ /^(.*\Q$sep\E)?$sampleRE/) {
        # found a valid sample filename in archive contents
        if ($packed eq "Q") {
            # extract sample path from archive contents
            # !!! This is only working for the packma test (only file-ITW viruses) !!!
            if ($content =~ /^.*((FILE|MALW)\Q$sep\E.*)\Q$sep\E($sampleRE).*$/
               and defined($1) and $1 ne "") {
                $newPath .= $sep . $1 . $sep . $3;
            }
        } else {
            # discard everything but the first valid filename from the content
            $content =~ s/^(.*\Q$sep\E)?($sampleRE).*$/$2/;
            $newPath .= $sep . $content;
        }
    }

    return $newPath;
}


# This one is used for products, that print a '->' (or similar chars)
# after the archive filename and then print the archive contents.
#sub modifyPackedCore {
#    my $pack = $_[0];
#    my $oldPath = $_[1];
#    my ($tempPath, $tempFile);
#    my $packFile = $packPrg{$pack}{"filename"};
#    # only proceed for the correct packer
#    # "-1" is used to signal a non-match, as it newver can be a valid
#    # absolute filepath
#    return "-1" unless ($oldPath =~ /(\\|\/)\Q$packFile\E(\\|\/)?/);
#
#    if ($debug > 1) {
#        &$logRef("found '".$packFile."'  (packer ".$pack.")\n", 5);
#    }
#    # split pathname into archive path and archive contents
#    # ATTENTION: Don't forget to escape chars in the regex for 'split'
### EXAMPLE:
#    ($tempPath, $tempFile) = split(/=>/, $oldPath, 2);
#
#    # for 'compressOnly' packers the archive file name is the directory and the
#    # files in the directory have the names of the archive contents
#    # Yes, this not an optimal solution ...
#    if ($tempPath =~ /\Q$packFile$sep\E$sampleRE$/) {
#        return $tempPath;
#    }
#
#    if (defined($tempFile) and $tempFile ne "") {
#        # find the last content containing a valid sample filename
#        # (there may be many nested contents in an archive)
### EXAMPLE:
#        if ($tempFile =~ /=>.+$/) {
### EXAMPLE:
#            my @content = split(/=>/, $tempFile);
#            # set content to a value that must be ignored later on, just in case
#            # no valid content is found
#            $tempFile = "(ERROR)";
#            for (my $i = $#content; $i >= 0; $i--) {
#                if ($debug > 1) {
#                   &$logRef("looking at index $i: '".$content[$i]."'\n", 6);
#                }
#                if ($content[$i] =~ /$sampleRE$/) {
#                    $tempFile = $content[$i];
#                    last;
#                }
#            }
#        }
#        # only process contents with a valid sample filename
#        # (to avoid any garbage appended to the filename)
#        if ($tempFile =~ /$sampleRE$/) {
#            # found a valid sample filename in archive contents
#            if ($packed eq "Q") {
#                # convert absolute paths into paths relative to testbed root
#                $tempFile =~ s/^[A-Z]:\Q$sep\E//;
#            } else {
#                # discard everything but the last filename from the content
#                #$tempFile =~ s/^(.*(\\|\/))?([^\\\/ ]+)$/$3/;
#                $tempFile =~ s/^(.*\Q$sep\E)?($sampleRE)$/$2/;
#            }
#            $tempPath .=  $sep . $tempFile;
#        }
#    }
#
#    return $tempPath;
#}


# cut of all additional characters appended to the filename
sub correctPath {
    my $path = $_[0];
    my $count;
    my $cutOff;
    # just to reset all positional variables ($1, $2, ...)
    $path =~ /.*/;
    # cut everything appended to a normalized filename (used in most testbeds)
    # since this regex is greedy it catches ALL matching filenames from
    # the end of the string to the first matching filename
    $path =~ s/($sampleRE)(.*)$/$1/i;
    if (defined($2) and ($2 ne "")) {
        $cutOff = $2;
    }
    if (defined($cutOff) and ($cutOff ne "")) {
        printf(CUT "line: %6d - cut from path: %s%s", $lineNr,
                   $cutOff, $EOL);
    }
    return $path;
}

# For these '->' products, there's also a function to correct the path
# for all other (= non-packed) testbeds
# cut of all additional characters appended to the filename
#sub correctPath {
#    my $path = $_[0];
#    my $cutOff;
### EXAMPLE:
#    ($path, $cutOff) = split(/->/, $path, 2);
#    if (defined($cutOff) and ($cutOff ne "")) {
#        printf(CUT "line: %6d - cut from path: %s%s", $lineNr,
### EXAMPLE:
#                   "->".$cutOff, $EOL);
#    }
#    return $path;
#}


# Check the virusname if it's really an identification or just a
# generic string. Use $VirusNameUI as name for unreliable detected samples.
# Eicar-testfiles must be sorted out too. Set $testFile to '1' to do that.
sub detectUnreliable {
    my $name = $_[0];
    my $upName = uc($name);
    # the EICAR testfile is not counted as virus
### EXAMPLE:
    if ($upName eq "EICAR_TEST_FILE") {
        $testFile = 1;
    }
### EXAMPLE:
    elsif ($upName eq "VBS.I-WORM" or $upName =~ /TYPE_.+/) {
        $name = $VirusNameUI;
    }
    # replace whitespaces with underscores
    $name =~ s/ /_/g;
    return $name;
}